"""
utils.py - Bot utilities

February 2020, Lewis Gaul
"""

__all__ = (
    "BOT_NAME",
    "USER_NAMES",
    "USER_NAMES_FILE",
    "Matchup",
    "get_matchups",
    "get_message",
    "get_highscore_times",
    "send_group_message",
    "send_message",
    "send_new_best_message",
    "send_myself_message",
    "set_bot_access_token",
    "set_user_nickname",
    "user_from_email",
)

import collections
import json
import logging
import pathlib
from typing import Dict, Iterable, List, Optional, Tuple

import requests
from requests_toolbelt import MultipartEncoder

from minegauler.shared import highscores as hs


logger = logging.getLogger(__name__)


USER_NAMES = dict()
USER_NAMES_FILE = pathlib.Path(__file__).parent / "users.json"

BOT_NAME = "minegaulerbot"
_BOT_ACCESS_TOKEN = ""
_BOT_EMAIL = f"{BOT_NAME}@webex.bot"
_WEBEX_GROUP_ROOM_ID = (
    "Y2lzY29zcGFyazovL3VzL1JPT00vNzYyNjI4NTAtMzg3Ni0xMWVhLTlhM2ItODMyNzMyZDlkZTg3"
)


# ------------------------------------------------------------------------------
# External
# ------------------------------------------------------------------------------


def set_bot_access_token(token: str) -> None:
    global _BOT_ACCESS_TOKEN
    _BOT_ACCESS_TOKEN = token


def get_message(msg_id: str) -> str:
    response = requests.get(
        f"https://api.ciscospark.com/v1/messages/{msg_id}",
        headers={"Authorization": f"Bearer {_BOT_ACCESS_TOKEN}"},
    )
    response.raise_for_status()
    return response.json()["text"]


def send_message(
    room_id: str, text: str, *, is_person_id=False, markdown=False
) -> requests.Response:
    logger.debug(
        "Sending message to %s:\n%s", "person" if is_person_id else "room", text
    )
    id_field = "toPersonId" if is_person_id else "roomId"
    text_field = "markdown" if markdown else "text"
    multipart = MultipartEncoder({text_field: text, id_field: room_id})
    response = requests.post(
        "https://api.ciscospark.com/v1/messages",
        data=multipart,
        headers={
            "Authorization": f"Bearer {_BOT_ACCESS_TOKEN}",
            "Content-Type": multipart.content_type,
        },
    )
    response.raise_for_status()
    return response


def send_myself_message(text: str) -> requests.Response:
    return send_message(_get_my_id(), text, is_person_id=True)


def send_group_message(text: str) -> requests.Response:
    return send_message(_WEBEX_GROUP_ROOM_ID, text)


def send_new_best_message(h: hs.HighscoreStruct) -> None:
    """Send a group message when a new personal time record is set."""
    if h.difficulty == "B":
        diff = "beginner"
    elif h.difficulty == "I":
        diff = "intermediate"
    elif h.difficulty == "E":
        diff = "expert"
    elif h.difficulty == "M":
        diff = "master"
    else:
        assert False
    drag_select = "on" if h.drag_select else "off"
    send_group_message(
        f"New personal record of {h.elapsed:.2f} set by {h.name} on {diff}!\n"
        f"Settings: drag-select={drag_select}, per-cell={h.per_cell}"
    )


def user_from_email(email: str) -> str:
    return email.split("@", maxsplit=1)[0]


def user_to_email(user: str) -> str:
    return f"{user}@cisco.com"


def tag_user(user: str) -> str:
    return f"<@personEmail:{user_to_email(user)}|{user}>"


def set_user_nickname(user: str, nickname: str) -> None:
    USER_NAMES[user] = nickname
    with open(USER_NAMES_FILE, "w") as f:
        json.dump(USER_NAMES, f)


def get_highscore_times(
    difficulty: Optional[str],
    drag_select: Optional[bool] = None,
    per_cell: Optional[int] = None,
    users: Optional[Iterable[str]] = None,
) -> List[Tuple[str, float]]:
    if users is None:
        users = USER_NAMES.values()

    if difficulty in ["beginner", "intermediate", "expert", "master"]:
        highscores = hs.filter_and_sort(
            hs.get_highscores(
                hs.HighscoresDatabases.REMOTE,
                difficulty=difficulty[0],
                drag_select=drag_select,
                per_cell=per_cell,
            )
        )
        times = {h.name: h.elapsed for h in highscores if h.name in users}
    else:
        assert difficulty is None
        times = dict.fromkeys(users, 0)
        for diff in ["beginner", "intermediate", "expert"]:
            highscores = hs.filter_and_sort(
                hs.get_highscores(
                    hs.HighscoresDatabases.REMOTE,
                    difficulty=diff[0],
                    drag_select=drag_select,
                    per_cell=per_cell,
                )
            )
            highscores = {h.name: h.elapsed for h in highscores if h.name in times}
            for name in list(times):
                if name in highscores:
                    times[name] += highscores[name]
                else:
                    times[name] += 1000

    return sorted(times.items(), key=lambda x: x[1])


Matchup = collections.namedtuple("Matchup", "user1, time1, user2, time2, percent")


def get_matchups(
    times: Iterable[Tuple[str, float]], include_users: Optional[Iterable[str]] = None
) -> List[Matchup]:
    times = sorted(times, key=lambda x: x[1], reverse=True)
    matchups = set()
    while times:
        # Avoid repeating matchups or comparing users against themselves.
        user1, time1 = times.pop()
        for user2, time2 in times:
            if (
                include_users
                and user1 not in include_users
                and user2 not in include_users
            ):
                continue
            assert time2 >= time1
            percent = 100 * (time2 - time1) / time1
            matchups.add(Matchup(user1, time1, user2, time2, percent))

    return sorted(matchups, key=lambda x: x.percent)


# ------------------------------------------------------------------------------
# Internal
# ------------------------------------------------------------------------------


def _strbool(b: bool) -> str:
    return "True" if b else "False"


def _get_person_id(name_or_email: str) -> str:
    if "@" in name_or_email:
        params = {"email": name_or_email}
    else:
        params = {"displayName": name_or_email}
    response = requests.get(
        f"https://api.ciscospark.com/v1/people",
        params=params,
        headers={"Authorization": f"Bearer {_BOT_ACCESS_TOKEN}"},
    )
    response.raise_for_status()
    return response.json()["items"][0]["id"]


def _get_my_id() -> str:
    return _get_person_id("Lewis Gaul")


def _get_bot_id() -> str:
    return _get_person_id("Minegauler Bot")



"""
api.py - API on to the core, providing all information needed by frontends

September 2019, Lewis Gaul

Exports:
TODO
"""

__all__ = (
    "AbstractController",
    "AbstractListener",
    "AbstractSwitchingController",
    "Caller",
    "EndedGameInfo",
)

import abc
import logging
import os
from typing import Callable, Dict, Iterable, List, Optional

import attr

from ..core import board as brd
from ..shared import utils
from ..types import CellContentsType, GameState, UIMode
from ..typing import Coord_T


@attr.attrs(auto_attribs=True)
class GameInfo:
    """General information about a game."""

    @attr.attrs(auto_attribs=True)
    class FinishedInfo:
        start_time: float
        elapsed: float
        bbbv: int
        rem_bbbv: int
        bbbvps: float
        prop_complete: float
        prop_flagging: float

    game_state: GameState
    x_size: int
    y_size: int
    mines: int
    per_cell: int
    finished_info: Optional[FinishedInfo] = None


@attr.attrs(auto_attribs=True)
class EndedGameInfo:
    """Information about a finished game."""

    game_state: GameState
    difficulty: str
    per_cell: int
    start_time: float
    elapsed: float
    bbbv: int
    flagging: float
    minefield_known: bool


class AbstractListener(metaclass=abc.ABCMeta):
    """
    An abstract class outlining methods that should be implemented to receive
    updates on changes to state. Instances of a concrete implementation can
    then be registered to listen for callbacks.
    """

    @abc.abstractmethod
    def reset(self) -> None:
        """
        Called to indicate the state should be reset.
        """
        return NotImplemented

    @abc.abstractmethod
    def resize_minefield(self, x_size: int, y_size: int) -> None:
        """
        Called to indicate the board is being changed.

        :param x_size:
            The number of columns.
        :param y_size:
            The number of rows.
        """
        return NotImplemented

    @abc.abstractmethod
    def set_mines(self, mines: int) -> None:
        """
        Called to indicate the number of mines at reset has changed.
        """
        return NotImplemented

    @abc.abstractmethod
    def update_cells(self, cell_updates: Dict[Coord_T, CellContentsType]) -> None:
        """
        Called when one or more cells were updated.

        :param cell_updates:
            Mapping of coordinates that were changed to the new cell state.
        """
        return NotImplemented

    @abc.abstractmethod
    def update_game_state(self, game_state: GameState) -> None:
        """
        Called when the game state changes.

        :param game_state:
            The new game state.
        """
        return NotImplemented

    @abc.abstractmethod
    def update_mines_remaining(self, mines_remaining: int) -> None:
        """
        Called when the number of mines remaining changes.

        :param mines_remaining:
            The new number of mines remaining.
        """
        return NotImplemented

    @abc.abstractmethod
    def handle_finished_game(self, info: EndedGameInfo) -> None:
        """
        Called once when a game ends.

        :param info:
            A store of end-game information.
        """
        return NotImplemented

    @abc.abstractmethod
    def switch_mode(self, mode: UIMode) -> None:
        """
        Called to indicate the game mode has change.

        :param mode:
            The mode to change to.
        """
        return NotImplemented

    @abc.abstractmethod
    def handle_exception(self, method: str, exc: Exception) -> None:
        """
        Called if an exception occurs when calling any of the other methods on
        the class, to allow the implementer of the class to handle (e.g. log)
        any errors.

        :param method:
            The method that the exception occurred in.
        :param exc:
            The caught exception.
        """
        return NotImplemented


class Caller(AbstractListener):
    """
    Pass on calls to registered listeners.
    """

    _count: int = 0

    def __init__(self, listeners: Iterable[AbstractListener] = None):
        """
        Create the implementation for all

        :param listeners:
        """
        self._listeners: List[AbstractListener] = list(listeners) if listeners else []
        self._id: int = self._count
        self._logger = logging.getLogger(f"{__name__}.{type(self).__name__}{self._id}")

        self.__class__._count += 1

        # Do the method wrapping here because we need the registered listeners.
        for method in AbstractListener.__abstractmethods__:
            setattr(self, method, self._call_registered(method))

    def register_listener(self, listener: AbstractListener) -> None:
        """
        Register a listener to receive updates from the controller.

        :param listener:
            An AbstractListener instance to register.
        """
        self._listeners.append(listener)

    def unregister_listener(self, listener: AbstractListener) -> None:
        """
        Unregister a listener to receive updates from the controller.

        Does nothing if not registered.

        :param listener:
            An AbstractListener instance to unregister.
        """
        try:
            self._listeners.remove(listener)
        except ValueError:
            self._logger.debug("Listener not registered - nothing to do")

    def _call_registered(self, func: str) -> Callable:
        """
        Decorator to call all registered listeners.

        :param func:
            The name of the method to decorate.
        :return:
            The decorated version of the method.
        """
        if not hasattr(self, func + "_orig"):
            setattr(self, func + "_orig", getattr(self, func))

        def wrapped(*args, **kwargs):
            getattr(self, func + "_orig")(*args, **kwargs)
            for listener in self._listeners:
                try:
                    getattr(listener, func)(*args, **kwargs)
                except Exception as e:
                    self._logger.warning(
                        f"Error ocurred calling {func}() on {listener}"
                    )
                    listener.handle_exception(func, e)

        return wrapped

    def reset(self) -> None:
        """
        Called to indicate the state should be reset.
        """
        self._logger.debug("Calling reset()")

    def resize_minefield(self, x_size: int, y_size: int) -> None:
        """
        Called to indicate the board shape has changed.

        :param x_size:
            The number of rows.
        :param y_size:
            The number of columns.
        """
        self._logger.debug(f"Calling resize_minefield() with {x_size}, {y_size}")

    def set_mines(self, mines: int) -> None:
        """
        Called to indicate the number of mines at reset has changed.
        """
        self._logger.debug(f"Calling set_mines() with {mines}")

    def update_cells(self, cell_updates: Dict[Coord_T, CellContentsType]) -> None:
        """
        Called when one or more cells were updated.

        :param cell_updates:
            Mapping of coordinates that were changed to the new cell state.
        """
        self._logger.debug(
            f"Calling update_cells() with {len(cell_updates)} updated cells"
        )

    def update_game_state(self, game_state: GameState) -> None:
        """
        Called when the game state changes.

        :param game_state:
            The new game state.
        """
        self._logger.debug(f"Calling update_game_state() with {game_state}")

    def update_mines_remaining(self, mines_remaining: int) -> None:
        """
        Called when the number of mines remaining changes.

        :param mines_remaining:
            The new number of mines remaining.
        """
        self._logger.debug(f"Calling update_mines_remaining() with {mines_remaining}")

    def handle_finished_game(self, info: EndedGameInfo) -> None:
        """
        Called once when a game ends.

        :param info:
            A store of end-game information.
        """
        self._logger.debug(f"Calling handle_finished_game() with {info}")

    def switch_mode(self, mode: UIMode) -> None:
        """
        Called to indicate the game mode has change.

        :param mode:
            The mode to change to.
        """
        self._logger.debug(f"Calling switch_mode() with {mode}")

    def handle_exception(self, method: str, exc: Exception) -> None:
        """
        Not used in this class - provided only to satisfy the ABC.
        """
        return NotImplemented


class AbstractController(metaclass=abc.ABCMeta):
    """
    Abstract controller base class. Listeners can be registered for receiving
    updates.
    """

    def __init__(self, opts: utils.GameOptsStruct, *, notif: Optional[Caller] = None):
        self._opts = utils.GameOptsStruct.from_structs(opts)
        # The registered functions to be called with updates.
        self._notif: Caller = notif if notif else Caller()
        self._logger = logging.getLogger(
            ".".join([type(self).__module__, type(self).__name__])
        )

    def register_listener(self, listener: AbstractListener) -> None:
        """
        Register a listener to receive updates from the controller.

        :param listener:
            An AbstractListener instance to register.
        """
        self._logger.info(
            "Registering listener: %s.%s",
            type(listener).__module__,
            type(listener).__name__,
        )
        self._notif.register_listener(listener)
        # TODO Implement in subclass to update with current state.

    def unregister_listener(self, listener: AbstractListener) -> None:
        """
        Unregister a listener to receive updates from the controller.

        Does nothing if not registered.

        :param listener:
            An AbstractListener instance to unregister.
        """
        self._logger.info(
            "Unregistering listener: %s.%s",
            type(listener).__module__,
            type(listener).__name__,
        )
        self._notif.unregister_listener(listener)

    @property
    @abc.abstractmethod
    def board(self) -> brd.Board:
        return NotImplemented

    @abc.abstractmethod
    def get_game_info(self) -> GameInfo:
        """Get information about the current game."""
        return NotImplemented

    def get_game_options(self) -> utils.GameOptsStruct:
        return self._opts

    # --------------------------------------------------------------------------
    # Methods triggered by user interaction
    # --------------------------------------------------------------------------
    @abc.abstractmethod
    def new_game(self) -> None:
        """
        Create a new game, refresh the board state.
        """
        self._logger.info("New game requested, refreshing the board")

    @abc.abstractmethod
    def restart_game(self) -> None:
        """
        Restart the current game, refresh the board state.
        """
        self._logger.info("Restart game requested, refreshing the board")

    @abc.abstractmethod
    def select_cell(self, coord: Coord_T) -> None:
        """
        Select a cell for a regular click.
        """
        self._logger.debug("Cell %s selected", coord)

    @abc.abstractmethod
    def flag_cell(self, coord: Coord_T, *, flag_only: bool = False) -> None:
        """
        Select a cell for flagging.
        """
        self._logger.debug("Cell %s selected for flagging", coord)

    @abc.abstractmethod
    def chord_on_cell(self, coord: Coord_T) -> None:
        """
        Select a cell for chording.
        """
        self._logger.debug("Cell %s selected for chording", coord)

    @abc.abstractmethod
    def remove_cell_flags(self, coord: Coord_T) -> None:
        """
        Remove flags in a cell, if any.
        """
        self._logger.debug("Flags in cell %s being removed", coord)

    @abc.abstractmethod
    def resize_board(self, x_size: int, y_size: int, mines: int) -> None:
        """
        Resize the board and/or change the number of mines.
        """
        self._logger.info(
            "Resizing the board to %sx%s with %s mines", x_size, y_size, mines
        )

    @abc.abstractmethod
    def set_first_success(self, value: bool) -> None:
        """
        Set whether the first click should be a guaranteed success.
        """
        self._logger.debug("Setting first success to %s", value)

    @abc.abstractmethod
    def set_per_cell(self, value: int) -> None:
        """
        Set the maximum number of mines per cell.
        """
        self._logger.debug("Setting per cell to %s", value)

    @abc.abstractmethod
    def save_current_minefield(self, file: os.PathLike) -> None:
        """
        Save the current minefield to file.

        :param file:
            The location of the file to save to. Should have the extension
            ".mgb".
        """
        self._logger.debug("Saving current minefield to file: %s", file)

    @abc.abstractmethod
    def load_minefield(self, file: os.PathLike) -> None:
        """
        Load a minefield from file.

        :param file:
            The location of the file to load from. Should have the extension
            ".mgb".
        """
        self._logger.debug("Loading minefield from file: %s", file)


class AbstractSwitchingController(AbstractController):
    """An abstract base class for controllers that support switching mode."""

    @abc.abstractmethod
    def switch_mode(self, mode: UIMode) -> None:
        """
        Switch the mode of the UI, e.g. into 'create' mode.
        """
        self._logger.info("Requested switch to mode %s", mode)



"""
__main__.py - Server entry-point

January 2020, Lewis Gaul
"""

import argparse
import logging
import os
import sys

import attr
from flask import Flask, abort, jsonify, redirect, request

from minegauler.shared import highscores as hs
from server import bot
from server.utils import is_highscore_new_best

from . import get_new_highscore_hooks


logger = logging.getLogger(__name__)

app = Flask(__name__)


# ------------------------------------------------------------------------------
# REST API
# ------------------------------------------------------------------------------


@app.route("/api/v1/highscore", methods=["POST"])
def api_v1_highscore():
    """
    Notification of a new highscore being set.

    Perform any desired handling for each highscore (e.g. usage logging), and
    also perform special handling for new records (e.g. add to the remote DB).
    """
    data = request.get_json()
    # verify_highscore(data)  TODO
    highscore = hs.HighscoreStruct.from_dict(data)
    logger.debug("POST highscore: %s", highscore)

    new_best = is_highscore_new_best(highscore)
    if new_best is None:
        logger.debug("Not a new best, ignoring the highscore")
        return "", 200

    try:
        hs.RemoteHighscoresDB().insert_highscore(highscore)
    except hs.DBConnectionError as e:
        logger.exception("Failed to insert highscore into remote DB")
        # TODO: I want to know if this is hit!
        return str(e), 503

    for func in get_new_highscore_hooks():
        try:
            func(highscore)
        except BaseException:
            logger.exception(f"Error in 'new highscore' hook {func.__name__}()")

    return "", 200


@app.route("/api/v1/highscores", methods=["GET"])
def api_v1_highscores():
    """Provide a REST API to get highscores from the DB."""
    logger.debug("GET highscores with args: %s", dict(request.args))
    difficulty = request.args.get("difficulty")
    per_cell = request.args.get("per_cell")
    if per_cell:
        per_cell = int(per_cell)
    drag_select = request.args.get("drag_select")
    if drag_select:
        drag_select = bool(int(drag_select))
    name = request.args.get("name")
    return jsonify(
        [
            attr.asdict(h)
            for h in hs.get_highscores(
                hs.HighscoresDatabases.REMOTE,
                drag_select=drag_select,
                per_cell=per_cell,
                difficulty=difficulty,
                name=name,
            )
        ]
    )


@app.route("/api/v1/highscores/ranks", methods=["GET"])
def api_v1_highscores_ranks():
    """Provide a REST API to get highscores from the DB."""
    logger.debug("GET highscores with args: %s", dict(request.args))
    difficulty = request.args.get("difficulty")
    per_cell = request.args.get("per_cell")
    drag_select = request.args.get("drag_select")
    if not difficulty or not per_cell or not drag_select:
        abort(404)
    per_cell = int(per_cell)
    drag_select = bool(int(drag_select))
    return jsonify(
        [
            attr.asdict(h)
            for h in hs.filter_and_sort(
                hs.get_highscores(
                    hs.HighscoresDatabases.REMOTE,
                    drag_select=drag_select,
                    per_cell=per_cell,
                    difficulty=difficulty,
                )
            )
        ]
    )


# ------------------------------------------------------------------------------
# Webpage serving
# ------------------------------------------------------------------------------


@app.route("/")
def index():
    return redirect("https://www.lewisgaul.co.uk/minegauler", 302)


@app.route("/highscores")
def highscores():
    return api_v1_highscores()


# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------


def _parse_args(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("--port", "-p", type=int, help="Override the default port")
    parser.add_argument("--bot", action="store_true", help="Handle bot messages")
    parser.add_argument("--dev", action="store_true", help="Run in development mode")
    return parser.parse_args(argv)


def main(argv):
    if "SQL_DB_PASSWORD" not in os.environ:
        logger.error("No 'SQL_DB_PASSWORD' env var set")
        sys.exit(1)

    args = _parse_args(argv)

    logging.basicConfig(
        filename="server.log",
        level=logging.DEBUG,
        format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
    )

    if args.bot:
        bot.init_route_handling(app)

    logger.info("Starting up")
    if args.dev:
        os.environ["FLASK_ENV"] = "development"
        app.run(debug=True, port=args.port)
    else:
        from waitress import serve

        serve(app, port=args.port if args.port else 80)


main(sys.argv[1:])



"""
highscores.py - Highscores handling

December 2019, Felix Gaul
"""

__all__ = (
    "HighscoreSettingsStruct",
    "HighscoreStruct",
    "HighscoresDatabases",
    "filter_and_sort",
    "get_highscores",
    "insert_highscore",
)

import abc
import enum
import logging
import os
import sqlite3
import threading
from textwrap import dedent
from typing import Dict, Iterable, List, Optional, Tuple

import attr
import mysql.connector
import requests

from .. import ROOT_DIR
from ..utils import StructConstructorMixin
from . import utils


logger = logging.getLogger(__name__)

_REMOTE_POST_URL = "http://minegauler.lewisgaul.co.uk/api/v1/highscore"


@attr.attrs(auto_attribs=True)
class HighscoreSettingsStruct(StructConstructorMixin):
    """A set of highscore settings."""

    difficulty: str
    per_cell: int
    drag_select: bool

    def __getitem__(self, item):
        return getattr(self, item)

    @classmethod
    def get_default(cls) -> "HighscoreSettingsStruct":
        return cls("B", 1, False)


@attr.attrs(auto_attribs=True)
class HighscoreStruct(HighscoreSettingsStruct):
    """A single highscore."""

    name: str
    timestamp: int
    elapsed: float
    bbbv: int
    bbbvps: float
    flagging: float

    @classmethod
    def from_sqlite_row(cls, cursor: sqlite3.Cursor, row: Tuple) -> "HighscoreStruct":
        """Create an instance from an SQLite row."""
        return cls(**{col[0]: row[i] for i, col in enumerate(cursor.description)})


_highscore_fields = attr.fields_dict(HighscoreStruct).keys()


class DBConnectionError(Exception):
    """Unable to connect to a database."""


class AbstractHighscoresDB(abc.ABC):
    """Abstract base class for a highscores database."""

    @property
    @abc.abstractmethod
    def conn(self):
        """The active database connection."""
        return None

    @abc.abstractmethod
    def get_highscores(
        self,
        *,
        difficulty: Optional[str] = None,
        per_cell: Optional[int] = None,
        drag_select: Optional[bool] = None,
        name: Optional[str] = None,
    ) -> Iterable[HighscoreStruct]:
        """Fetch highscores from the database using the given filters."""
        logger.debug("%s: Getting highscores", type(self).__name__)

    @abc.abstractmethod
    def insert_highscore(self, highscore: HighscoreStruct) -> None:
        """Insert a single highscore into the database."""
        logger.debug(
            "%s: Inserting highscore into DB: %s", type(self).__name__, highscore
        )

    def execute(self, cmd: str, params: Tuple = (), *, commit=False, **cursor_args):
        """
        Execute a command on the database.

        :param cmd:
            The command to execute.
        :param params:
            Parameters to pass to the command.
        :param commit:
            Whether to do a commit after executing the command.
        :param cursor_args:
            Keyword arguments to pass on when creating the DB cursor.
        :raise DBConnectionError:
            If executing the command fails due to loss of connection.
        """
        cursor = self.conn.cursor(**cursor_args)
        logger.debug(
            "%s: Executing command %r with params: %s", type(self).__name__, cmd, params
        )
        cursor.execute(cmd, params)
        if commit:
            self.conn.commit()
        return cursor


class _SQLMixin:
    """A mixin for SQL-like highscores databases."""

    _TABLE_NAME = "highscores"
    _CREATE_TABLE_SQL = dedent(
        f"""\
        CREATE TABLE IF NOT EXISTS {_TABLE_NAME} (
            id INTEGER PRIMARY KEY,
            difficulty VARCHAR(1) NOT NULL,
            per_cell INTEGER NOT NULL,
            drag_select INTEGER NOT NULL,
            name VARCHAR(20) NOT NULL,
            timestamp INTEGER NOT NULL,
            elapsed REAL NOT NULL,
            bbbv INTEGER NOT NULL,
            bbbvps REAL NOT NULL,
            flagging REAL NOT NULL
        )"""
    )

    def _get_select_highscores_sql(
        self,
        *,
        difficulty: Optional[str] = None,
        per_cell: Optional[int] = None,
        drag_select: Optional[bool] = None,
        name: Optional[str] = None,
    ) -> str:
        """Get the SQL command to get/select highscores from a DB."""
        conditions = []
        if difficulty is not None:
            conditions.append(f"difficulty='{difficulty}'")
        if per_cell is not None:
            conditions.append(f"per_cell={per_cell}")
        if drag_select is not None:
            conditions.append(f"drag_select={drag_select:d}")
        if name is not None:
            conditions.append(f"name='{name}'")
        return "SELECT {} FROM highscores {} ORDER BY elapsed ASC".format(
            ", ".join(_highscore_fields),
            "WHERE " + " AND ".join(conditions) if conditions else "",
        )

    def _get_insert_highscore_sql(self, format="%s") -> str:
        """Get the SQL command to insert a highscore into a DB."""
        return "INSERT INTO highscores ({}) " "VALUES ({})".format(
            ", ".join(_highscore_fields), ", ".join(format for _ in _highscore_fields)
        )


class LocalHighscoresDB(_SQLMixin, AbstractHighscoresDB):
    """Database of local highscores."""

    _DB_FILE = ROOT_DIR / "data" / "highscores.db"

    def __init__(self):
        if os.path.exists(self._DB_FILE):
            self._conn = sqlite3.connect(str(self._DB_FILE))
        else:
            os.makedirs(self._DB_FILE.parent, exist_ok=True)
            self._conn = sqlite3.connect(str(self._DB_FILE))

            cursor = self._conn.cursor()
            cursor.execute(self._CREATE_TABLE_SQL)
        self._conn.row_factory = HighscoreStruct.from_sqlite_row

    @property
    def conn(self) -> sqlite3.Connection:
        return self._conn

    def get_highscores(
        self,
        *,
        difficulty: Optional[str] = None,
        per_cell: Optional[int] = None,
        drag_select: Optional[bool] = None,
        name: Optional[str] = None,
    ) -> Iterable[HighscoreStruct]:
        super().get_highscores(
            difficulty=difficulty, per_cell=per_cell, drag_select=drag_select, name=name
        )
        cursor = self.execute(
            self._get_select_highscores_sql(
                difficulty=difficulty,
                per_cell=per_cell,
                drag_select=drag_select,
                name=name,
            )
        )
        return cursor.fetchall()

    def insert_highscore(self, highscore: HighscoreStruct) -> None:
        super().insert_highscore(highscore)
        self.execute(
            self._get_insert_highscore_sql(format="?"),
            attr.astuple(highscore),
            commit=True,
        )

    def execute(
        self, cmd: str, params: Tuple = (), *, commit=False, **cursor_args
    ) -> sqlite3.Cursor:
        return super().execute(cmd, params, commit=commit, **cursor_args)


class RemoteHighscoresDB(_SQLMixin, AbstractHighscoresDB):
    """Remote highscores database."""

    _USER = "admin"
    _PASSWORD = os.environ.get("SQL_DB_PASSWORD")
    _HOST = "minegauler-highscores.cb4tvkuqujyi.eu-west-2.rds.amazonaws.com"
    _DB_NAME = "minegauler"
    _TABLE_NAME = "highscores"

    _cached_conn: Optional[mysql.connector.MySQLConnection] = None

    def __init__(self):
        """
        :raise DBConnectionError:
            If connecting to the DB fails for any reason.
        """
        if not self._cached_conn:
            logger.info("Initialising connection to remote highscores DB")
            try:
                self._cached_conn = mysql.connector.connect(
                    user=self._USER,
                    password=self._PASSWORD,
                    host=self._HOST,
                    database=self._DB_NAME,
                )
            except mysql.connector.Error as e:
                raise DBConnectionError(
                    "Unable to connect to remote highscores database"
                ) from e
        self._conn = self._cached_conn

    @property
    def conn(self) -> mysql.connector.MySQLConnection:
        return self._conn

    def get_highscores(
        self,
        *,
        difficulty: Optional[str] = None,
        per_cell: Optional[int] = None,
        drag_select: Optional[bool] = None,
        name: Optional[str] = None,
    ) -> Iterable[HighscoreStruct]:
        super().get_highscores(
            difficulty=difficulty, per_cell=per_cell, drag_select=drag_select, name=name
        )
        cursor = self.execute(
            self._get_select_highscores_sql(
                difficulty=difficulty,
                per_cell=per_cell,
                drag_select=drag_select,
                name=name,
            ),
            dictionary=True,
        )
        return [HighscoreStruct(**r) for r in cursor.fetchall()]

    def insert_highscore(self, highscore: HighscoreStruct) -> None:
        super().insert_highscore(highscore)
        self.execute(
            self._get_insert_highscore_sql(), attr.astuple(highscore), commit=True
        )

    def execute(
        self, cmd: str, params: Tuple = (), *, commit=False, **cursor_args
    ) -> sqlite3.Cursor:
        try:
            return super().execute(cmd, params, commit=commit, **cursor_args)
        except mysql.connector.Error as e:
            raise DBConnectionError("Error occurred trying to execute command") from e


class HighscoresDatabases(enum.Enum):
    """An enum of highscores databases."""

    LOCAL = LocalHighscoresDB
    REMOTE = RemoteHighscoresDB

    def get_db_instance(self) -> AbstractHighscoresDB:
        return self.value()


def get_highscores(
    database=HighscoresDatabases.LOCAL,
    *,
    settings: Optional[HighscoreSettingsStruct] = None,
    difficulty: Optional[str] = None,
    per_cell: Optional[int] = None,
    drag_select: Optional[bool] = None,
    name: Optional[str] = None,
) -> Iterable[HighscoreStruct]:
    """
    Fetch highscores from a database.

    :param database:
        The database type to fetch from.
    :param settings:
        Optionally specify settings to filter by.
    :param difficulty:
        Optionally specify difficulty to filter by. Ignored if settings given.
    :param per_cell:
        Optionally specify per_cell to filter by. Ignored if settings given.
    :param drag_select:
        Optionally specify drag_select to filter by. Ignored if settings given.
    :param name:
        Optionally specify a name to filter by.
    """
    if settings is not None:
        difficulty = settings.difficulty
        per_cell = settings.per_cell
        drag_select = settings.drag_select
    return database.get_db_instance().get_highscores(
        difficulty=difficulty, per_cell=per_cell, drag_select=drag_select, name=name
    )


def insert_highscore(highscore: HighscoreStruct) -> None:
    """Insert a highscore into DBs."""
    LocalHighscoresDB().insert_highscore(highscore)

    def _post_catch_exc():
        try:
            _post_highscore_to_remote(highscore)
        except Exception:
            logger.exception("Failed to insert highscore into remote DB")

    threading.Thread(target=_post_catch_exc).start()


def filter_and_sort(
    highscores: Iterable[HighscoreStruct],
    sort_key: str = "time",
    filters: Dict[str, Optional[str]] = {},
) -> List[HighscoreStruct]:
    """
    Filter and sort an iterable of highscores.

    :param highscores:
        The iterable of highscores to filter and sort.
    :param sort_key:
        What to sort by.
    :param filters:
        What filters to apply.
    :return:
        A new iterable of highscores.
    """
    ret = []
    filters = {k: f for k, f in filters.items() if f}
    for hs in highscores:
        all_pass = True
        if "flagging" in filters:
            if (
                filters["flagging"] == "F"
                and not utils.is_flagging_threshold(hs.flagging)
                or filters["flagging"] == "NF"
                and utils.is_flagging_threshold(hs.flagging)
            ):
                all_pass = False
        if "name" in filters and filters["name"] != hs.name:
            all_pass = False
        if all_pass:
            # All filters satisfied.
            ret.append(hs)
    # Sort first by either time or 3bv/s, then by 3bv if there's a tie
    #  (higher 3bv higher for equal time, lower for equal 3bv/s).
    if sort_key == "time":
        ret.sort(key=lambda h: (h.elapsed, -h.bbbv))
    elif sort_key == "3bv/s":
        ret.sort(key=lambda h: (h.bbbvps, -h.bbbv), reverse=True)
    if "name" not in filters:
        # If no name filter, only include best highscore for each name.
        names = []
        i = 0
        while i < len(ret):
            hs = ret[i]
            name = hs["name"].lower()
            if name in names:
                ret.pop(i)
            else:
                names.append(name)
                i += 1
    return ret


def is_highscore_new_best(
    highscore: HighscoreStruct, all_highscores: Iterable[HighscoreStruct]
) -> Optional[str]:
    """
    Test to see if a new top highscore has been set.

    :param highscore:
        The highscore to check.
    :param all_highscores:
        The list of highscores to check against. May or may not include the
        highscore being checked.
    :return:
        If a new highscore was set, return which category it was set in. If not,
        return None.
    """
    all_highscores = list(all_highscores)
    top_time = filter_and_sort(all_highscores, "time", {"name": highscore.name})
    top_3bvps = filter_and_sort(all_highscores, "3bv/s", {"name": highscore.name})
    if not top_time or highscore.elapsed <= top_time[0].elapsed:
        return "time"
    elif not top_3bvps or highscore.bbbvps >= top_3bvps[0].bbbvps:
        return "3bv/s"
    else:
        return None


def _post_highscore_to_remote(highscore: HighscoreStruct):
    """Send a highscore to the remote server to be added to the remote DB."""
    logger.info("Posting highscore to remote")
    requests.post(_REMOTE_POST_URL, json=attr.asdict(highscore), timeout=5)



"""
__init__.py - Code shared between core and frontend

December 2019, Lewis Gaul
"""

__all__ = (
    "AllOptsStruct",
    "GameOptsStruct",
    "GUIOptsStruct",
    "HighscoreSettingsStruct",
    "HighscoreStruct",
    "get_difficulty",
    "highscores",
    "read_settings_from_file",
    "write_settings_to_file",
)

from . import highscores
from .highscores import HighscoreSettingsStruct, HighscoreStruct
from .utils import (
    AllOptsStruct,
    GameOptsStruct,
    GUIOptsStruct,
    get_difficulty,
    read_settings_from_file,
    write_settings_to_file,
)



"""
main_window.py - Base GUI application implementation

April 2018, Lewis Gaul

Exports:
BaseMainWindow
    Main window class.
    
MinegaulerGUI
    Minegauler main window class.
"""

__all__ = ("MinegaulerGUI",)

import logging
import os
import textwrap
import time as tm
import traceback
from typing import Callable, Dict, Mapping, Optional, Type

from PyQt5.QtCore import QSize, Qt, pyqtSignal
from PyQt5.QtGui import QFocusEvent, QFont, QIcon, QKeyEvent
from PyQt5.QtWidgets import (
    QAction,
    QActionGroup,
    QDialog,
    QFileDialog,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMainWindow,
    QMenu,
    QMenuBar,
    QPushButton,
    QSizePolicy,
    QSlider,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from .. import ROOT_DIR, shared
from ..core import api
from ..shared.highscores import HighscoreSettingsStruct, HighscoreStruct
from ..shared.utils import GUIOptsStruct, get_difficulty
from ..types import CellContentsType, CellImageType, GameState, UIMode
from ..typing import Coord_T
from . import highscores, minefield, panel, state, utils
from .utils import FILES_DIR


logger = logging.getLogger(__name__)

BOARD_DIR = ROOT_DIR / "boards"


class _BaseMainWindow(QMainWindow):
    """
    Base class for the application implementing the general layout.
    """

    BODY_FRAME_WIDTH = 10

    def __init__(self, parent: Optional[QWidget], title: Optional[str] = None):
        """
        :param title:
            The window title.
        """
        super().__init__(parent)
        self._menubar: QMenuBar = self.menuBar()
        self._game_menu: QMenu = self._menubar.addMenu("Game")
        self._opts_menu: QMenu = self._menubar.addMenu("Options")
        self._help_menu: QMenu = self._menubar.addMenu("Help")
        self._panel_frame: QFrame
        self._body_frame: QFrame
        self._footer_frame: QFrame
        self._panel_widget: Optional[QWidget]
        self._body_widget: Optional[QWidget]
        self._footer_widget: Optional[QWidget]
        self._icon: QIcon = QIcon(str(utils.IMG_DIR / "icon.ico"))
        self.setWindowTitle(title)
        self.setWindowIcon(self._icon)
        # Disable maximise button
        self.setWindowFlags(self.windowFlags() | Qt.CustomizeWindowHint)
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowMaximizeButtonHint)
        self._setup_ui()
        # Keep track of all non-modal subwindows that are open.
        self._open_subwindows: Dict[Type[QWidget], QWidget] = {}

    # --------------------------------------------------------------------------
    # UI setup
    # --------------------------------------------------------------------------
    def _setup_ui(self):
        """
        Set up the layout of the main window GUI.
        """
        # QMainWindow objects have a central widget to be set.
        central_widget = QWidget(self)
        central_widget.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.setCentralWidget(central_widget)
        vlayout = QVBoxLayout(central_widget)
        vlayout.setContentsMargins(0, 0, 0, 0)
        vlayout.setSpacing(0)
        # Top panel widget.
        self._panel_frame = QFrame(central_widget)
        self._panel_frame.setFrameShadow(QFrame.Sunken)
        self._panel_frame.setFrameShape(QFrame.Panel)
        self._panel_frame.setLineWidth(2)
        self._panel_frame.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Fixed)
        vlayout.addWidget(self._panel_frame)
        # Main body widget config - use horizontal layout for centre alignment.
        hstretch = QHBoxLayout()
        hstretch.addStretch()  # left-padding for centering
        self._body_frame = QFrame(central_widget)
        self._body_frame.setFrameShadow(QFrame.Raised)
        self._body_frame.setFrameShape(QFrame.Box)
        self._body_frame.setLineWidth(self.BODY_FRAME_WIDTH / 2)
        hstretch.addWidget(self._body_frame)
        hstretch.addStretch()  # right-padding for centering
        vlayout.addLayout(hstretch)
        # Name entry bar underneath the minefield
        self._footer_frame = QFrame(central_widget)
        vlayout.addWidget(self._footer_frame)

    def set_panel_widget(self, widget: QWidget) -> None:
        """
        Set the widget to occupy the top panel.

        Arguments:
        widget (QWidget)
            The widget instance to place in the top panel.
        """
        lyt = QVBoxLayout(self._panel_frame)
        lyt.setContentsMargins(0, 0, 0, 0)
        lyt.addWidget(widget)
        self._panel_widget = widget

    def set_body_widget(self, widget: QWidget) -> None:
        """
        Set the widget to occupy the main section of the GUI.

        Arguments:
        widget (QWidget)
            The widget instance to place in the body.
        """
        lyt = QVBoxLayout(self._body_frame)
        lyt.setContentsMargins(0, 0, 0, 0)
        lyt.addWidget(widget)
        self._body_widget = widget

    def set_footer_widget(self, widget: QWidget) -> None:
        """
        Set the widget to occupy the lower bar of the GUI.

        Arguments:
        widget (QWidget)
            The widget instance to place in the lower bar.
        """
        lyt = QVBoxLayout(self._footer_frame)
        lyt.setContentsMargins(0, 0, 0, 0)
        lyt.addWidget(widget)
        self._footer_widget = widget


class _MinegaulerGUIMeta(type(api.AbstractListener), type(_BaseMainWindow)):
    """Combined metaclass for the MinegaulerGUI class."""


class MinegaulerGUI(
    api.AbstractListener, _BaseMainWindow, metaclass=_MinegaulerGUIMeta
):
    """The main Minegauler GUI window."""

    def __init__(
        self, ctrlr: api.AbstractSwitchingController, initial_state: state.State
    ):
        """
        :param ctrlr:
            The core controller.
        :param initial_state:
            Initial application state.
        """
        super().__init__(None, "Minegauler")
        self._ctrlr: api.AbstractSwitchingController = ctrlr
        self._state: state.State = initial_state.deepcopy()

        self._create_menu_action: QAction
        self._diff_menu_actions: Dict[str, QAction] = dict()
        self._populate_menubars()
        self._menubar.setFixedHeight(self._menubar.sizeHint().height())
        self._panel_widget = panel.PanelWidget(self, self._state)
        self._mf_widget = minefield.MinefieldWidget(self, self._ctrlr, self._state)
        self.set_panel_widget(self._panel_widget)
        self.set_body_widget(self._mf_widget)
        self._name_entry_widget = _NameEntryBar(self, self._state.name)
        self.set_footer_widget(self._name_entry_widget)
        self._update_size()

        self._panel_widget.clicked.connect(self._ctrlr.new_game)
        self._mf_widget.at_risk_signal.connect(self._panel_widget.at_risk)
        self._mf_widget.no_risk_signal.connect(self._panel_widget.no_risk)
        self._mf_widget.size_changed.connect(self._update_size)
        self._name_entry_widget.name_updated_signal.connect(self._set_name)

    # ----------------------------------
    # Implemented abstractmethods
    # ----------------------------------
    def reset(self) -> None:
        """
        Called to indicate the GUI state should be soft-reset.

        This is distinct from a factory reset (settings are not changed).
        """
        self._panel_widget.reset()
        self._mf_widget.reset()

    def resize_minefield(self, x_size: int, y_size: int) -> None:
        """
        Called to indicate the board shape has changed.

        :param x_size:
            The number of columns.
        :param y_size:
            The number of rows.
        """
        self._state.x_size = x_size
        self._state.y_size = y_size
        self._mf_widget.reshape(x_size, y_size)
        # self._update_size()
        self._diff_menu_actions[self._state.difficulty].setChecked(True)

    def set_mines(self, mines: int) -> None:
        """
        Called to indicate the base number of mines has changed.

        :param mines:
            The number of mines.
        """
        self._state.mines = mines
        self._diff_menu_actions[self._state.difficulty].setChecked(True)

    def update_cells(self, cell_updates: Mapping[Coord_T, CellContentsType]) -> None:
        """
        Called to indicate some cells have changed state.

        :param cell_updates:
            A mapping of cell coordinates to their new state.
        """
        for c, state in cell_updates.items():
            self._mf_widget.set_cell_image(c, state)

    def update_game_state(self, game_state: GameState) -> None:
        """
        Called to indicate the game state has changed.

        :param game_state:
            The new game state.
        """
        self._state.game_status = game_state
        if game_state is GameState.READY:
            self._state.highscores_state.current_highscore = None
        self._panel_widget.update_game_state(game_state)

    def update_mines_remaining(self, mines_remaining: int) -> None:
        """
        Called to indicate the number of remaining mines has changed.

        :param mines_remaining:
            The remaining number of mines.
        """
        self._panel_widget.set_mines_counter(mines_remaining)

    def handle_finished_game(self, info: api.EndedGameInfo) -> None:
        """
        Called once when a game ends.

        :param info:
            A store of end-game information.
        """
        self._panel_widget.timer.stop()
        self._panel_widget.timer.set_time(int(info.elapsed + 1))
        # Store the highscore if the game was won.
        if (
            info.game_state is GameState.WON
            and info.difficulty != "C"
            and not info.minefield_known
        ):
            highscore = HighscoreStruct(
                difficulty=info.difficulty,
                per_cell=info.per_cell,
                timestamp=int(info.start_time),
                elapsed=info.elapsed,
                bbbv=info.bbbv,
                bbbvps=info.bbbv / info.elapsed,
                drag_select=self._state.drag_select,
                name=self._state.name,
                flagging=info.flagging,
            )
            try:
                shared.highscores.insert_highscore(highscore)
            except Exception:
                logger.exception("Error inserting highscore")
            self._state.highscores_state.current_highscore = highscore
            # Check whether to pop up the highscores window.
            # TODO: This is too slow...
            try:
                new_best = shared.highscores.is_highscore_new_best(
                    highscore, shared.highscores.get_highscores(settings=highscore)
                )
            except Exception:
                logger.exception("Error getting highscores")
            else:
                if new_best:
                    self.open_highscores_window(highscore, new_best)

    def switch_mode(self, mode: UIMode) -> None:
        """
        Called to indicate the game mode has change.

        :param mode:
            The mode to change to.
        """
        self._create_menu_action.setChecked(mode is UIMode.CREATE)

    def handle_exception(self, method: str, exc: Exception) -> None:
        logger.error(
            "Error occurred when calling %s() from backend:\n%s\n%s",
            method,
            "".join(traceback.format_exception(None, exc, exc.__traceback__)),
            exc,
        )
        raise RuntimeError(exc).with_traceback(exc.__traceback__)

    # --------------------------------------------------------------------------
    # Qt method overrides
    # --------------------------------------------------------------------------
    def sizeHint(self) -> QSize:
        width = self._body_frame.sizeHint().width()
        height = (
            self._menubar.sizeHint().height()
            + self._panel_frame.sizeHint().height()
            + self._body_frame.sizeHint().height()
            + self._name_entry_widget.sizeHint().height()
        )
        return QSize(width, height)

    def minimumSizeHint(self) -> QSize:
        width = self._panel_frame.minimumSizeHint().width()
        height = (
            self._menubar.sizeHint().height()
            + self._panel_frame.sizeHint().height()
            + 100
            + self._name_entry_widget.sizeHint().height()
        )
        return QSize(width, height)

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        self._name_entry_widget.clearFocus()

    # --------------------------------------------------------------------------
    # Other methods
    # --------------------------------------------------------------------------
    def _update_size(self):
        """Update the window size."""
        self.setMinimumSize(self.minimumSizeHint())
        self.setMaximumSize(self.sizeHint())
        self.resize(self.sizeHint())
        self.adjustSize()
        # print(self.sizeHint(), self.size())
        # print(self._menubar.sizeHint(), self._menubar.size())
        # print(self._panel_frame.sizeHint(), self._panel_frame.size())
        # print(self._body_frame.sizeHint(), self._body_frame.size())
        # print(self._mf_widget.sizeHint(), self._mf_widget.size())
        # print()

    def _populate_menubars(self) -> None:
        """Fill in the menubars."""

        # ----------
        # Game menu
        # ----------
        # New game (F2)
        new_game_act = self._game_menu.addAction("New game", self._ctrlr.new_game)
        new_game_act.setShortcut("F2")

        # Replay game (F3)
        replay_act = self._game_menu.addAction("Replay", self._ctrlr.restart_game)
        replay_act.setShortcut("F3")

        # Create board
        def switch_create_mode(checked: bool):
            mode = UIMode.CREATE if checked else UIMode.GAME
            self._state.ui_mode = mode
            self._ctrlr.switch_mode(mode)

        self._create_menu_action = create_act = QAction(
            "Create board",
            self,
            checkable=True,
            checked=self._state.ui_mode is UIMode.CREATE,
        )
        self._game_menu.addAction(create_act)
        create_act.triggered.connect(switch_create_mode)

        # Save board
        self._game_menu.addAction("Save board", self._open_save_board_modal)

        # Load board
        self._game_menu.addAction("Load board", self._open_load_board_modal)

        self._game_menu.addSeparator()

        # Current info (F4)
        info_act = self._game_menu.addAction(
            "Current game info", self._open_current_info_modal
        )
        info_act.setShortcut("F4")

        # Solver
        # - Probabilities (F5)
        # - Auto flag (Ctrl+F)
        # - Auto click (Ctrl+Enter)

        # Highscores (F6)
        highscores_act = self._game_menu.addAction(
            "Highscores", self.open_highscores_window
        )
        highscores_act.setShortcut("F6")

        # Stats (F7)

        self._game_menu.addSeparator()

        # Difficulty radiobuttons
        # - Beginner (b)
        # - Intermediate (i)
        # - Expert (e)
        # - Master (m)
        # - Custom (c)
        diff_group = QActionGroup(self)
        diff_group.setExclusive(True)
        for diff in ["Beginner", "Intermediate", "Expert", "Master", "Custom"]:
            diff_act = QAction(diff, diff_group, checkable=True)
            self._diff_menu_actions[diff[0]] = diff_act
            self._game_menu.addAction(diff_act)
            diff_act.id = diff[0]
            if diff_act.id == get_difficulty(
                self._state.x_size, self._state.y_size, self._state.mines
            ):
                diff_act.setChecked(True)
            diff_act.triggered.connect(
                lambda _: self._change_difficulty(diff_group.checkedAction().id)
            )
            diff_act.setShortcut(diff[0])

        self._game_menu.addSeparator()

        # Zoom
        self._game_menu.addAction("Button size", self._open_zoom_modal)

        # Styles
        # - Buttons
        # - Images
        # - Numbers
        def get_change_style_func(grp, style):
            def change_style():
                self._state.styles[grp] = style
                self._mf_widget.update_style(grp, style)

            return change_style

        styles_menu = QMenu("Styles", self)
        self._game_menu.addMenu(styles_menu)
        for img_group in [CellImageType.BUTTONS]:
            img_group_name = img_group.name.capitalize()
            submenu = QMenu(img_group_name, self)
            styles_menu.addMenu(submenu)
            group = QActionGroup(self)
            group.setExclusive(True)
            for folder in (utils.IMG_DIR / img_group_name).glob("*"):
                style = folder.name
                style_act = QAction(style, self, checkable=True)
                if style == self._state.styles[img_group]:
                    style_act.setChecked(True)
                group.addAction(style_act)
                style_act.triggered.connect(get_change_style_func(img_group, style))
                submenu.addAction(style_act)

        self._game_menu.addSeparator()

        # Exit (F4)
        self._game_menu.addAction("Exit", self.close, shortcut="Alt+F4")

        # ----------
        # Options menu
        # ----------
        # First-click success
        def toggle_first_success():
            new_val = not self._state.pending_first_success
            self._state.first_success = new_val
            self._ctrlr.set_first_success(new_val)

        first_act = QAction(
            "Safe start", self, checkable=True, checked=self._state.first_success
        )
        self._opts_menu.addAction(first_act)
        first_act.triggered.connect(toggle_first_success)

        # Drag select
        drag_act = self._opts_menu.addAction(
            "Drag select",
            lambda: setattr(
                self._state, "drag_select", not self._state.pending_drag_select
            ),
        )
        drag_act.setCheckable(True)
        drag_act.setChecked(self._state.drag_select)

        # Max mines per cell option
        def get_change_per_cell_func(n):
            def change_per_cell():
                self._state.per_cell = n
                self._ctrlr.set_per_cell(n)

            return change_per_cell

        per_cell_menu = self._opts_menu.addMenu("Max per cell")
        per_cell_group = QActionGroup(self)
        per_cell_group.setExclusive(True)
        for i in range(1, 4):
            action = QAction(str(i), self, checkable=True)
            per_cell_menu.addAction(action)
            per_cell_group.addAction(action)
            if self._state.per_cell == i:
                action.setChecked(True)
            action.triggered.connect(get_change_per_cell_func(i))

        # ----------
        # Help menu
        # ----------
        rules_act = QAction("Rules", self)
        self._help_menu.addAction(rules_act)
        rules_act.triggered.connect(
            lambda: self._open_text_popup("Rules", FILES_DIR / "rules.txt")
        )

        tips_act = QAction("Tips", self)
        self._help_menu.addAction(tips_act)
        tips_act.triggered.connect(
            lambda: self._open_text_popup("Tips", FILES_DIR / "tips.txt")
        )

        self._help_menu.addSeparator()

        about_act = QAction("About", self)
        self._help_menu.addAction(about_act)
        about_act.triggered.connect(
            lambda: self._open_text_popup("About", FILES_DIR / "about.txt")
        )
        about_act.setShortcut("F1")

    def _change_difficulty(self, id_: str) -> None:
        """
        Change the difficulty via the core controller.

        :param id_:
            The difficulty ('b', 'i', 'e', 'm' or 'c').
        """
        id_ = id_.upper()
        if id_ == "B":
            x, y, m = 8, 8, 10
        elif id_ == "I":
            x, y, m = 16, 16, 40
        elif id_ == "E":
            x, y, m = 30, 16, 99
        elif id_ == "M":
            x, y, m = 30, 30, 200
        elif id_ == "C":
            self._diff_menu_actions[self._state.difficulty].setChecked(True)
            logger.info("Opening popup window for selecting custom board")
            self._open_custom_board_modal()
            return
        else:
            raise ValueError(f"Unrecognised difficulty '{id_}'")

        self._ctrlr.resize_board(x_size=x, y_size=y, mines=m)

    def _set_name(self, name: str) -> None:
        self._state.name = name
        self._state.highscores_state.name_hint = name

    def _open_save_board_modal(self) -> None:
        if not (
            self._state.game_status.finished() or self._state.ui_mode is UIMode.CREATE
        ):
            return
        file, _ = QFileDialog.getSaveFileName(
            self,
            caption="Save board",
            directory=str(BOARD_DIR),
            filter="Minegauler boards (*.mgb)",
        )
        if not file:
            return
        elif not file.endswith(".mgb"):
            file += ".mgb"
        logger.info("Board requested to be saved at %s", file)
        try:
            self._ctrlr.save_current_minefield(file)
        except Exception:
            logger.exception("Error occurred trying to save minefield to file")

    def _open_load_board_modal(self) -> None:
        file, _ = QFileDialog.getOpenFileName(
            self,
            caption="Load board",
            directory=str(BOARD_DIR),
            filter="Minegauler boards (*.mgb)",
        )
        if not file:
            return
        logger.info("Board requested to be loaded from %s", file)
        try:
            self._ctrlr.load_minefield(file)
        except Exception:
            logger.exception("Error occurred trying to load minefield from file")

    def _open_current_info_modal(self) -> None:
        """Open a popup giving the current game info."""
        info = self._ctrlr.get_game_info()
        _CurrentInfoModal(self, info, self._state).show()

    def _open_custom_board_modal(self) -> None:
        """Open the modal popup to select the custom difficulty."""
        _CustomBoardModal(
            self,
            self._state.x_size,
            self._state.y_size,
            self._state.mines,
            self._ctrlr.resize_board,
        ).show()

    def _open_zoom_modal(self) -> None:
        """Open the popup to set the button size."""
        _ButtonSizeModal(self, self._state.btn_size, self.set_button_size).show()

    def _open_text_popup(self, title: str, file: os.PathLike):
        """Open a text popup window."""
        if title in self._open_subwindows:
            try:
                self._open_subwindows[title].close()
                # self._open_subwindows[title].setFocus()  # TODO: doesn't work!
                # return
            except:
                self._open_subwindows.pop(title)
        win = _TextPopup(self, title, file)
        win.show()
        self._open_subwindows[title] = win

    def get_gui_opts(self) -> GUIOptsStruct:
        return GUIOptsStruct.from_structs(self._state, self._state.pending_game_state)

    def open_highscores_window(
        self,
        settings: Optional[HighscoreSettingsStruct] = None,
        sort_by: Optional[str] = None,
    ) -> None:
        """
        Open the highscores window.

        :param settings:
            Optionally specify the highscore settings to display highscores
            for. By default the current frontend state is used.
        :param sort_by:
            Optionally specify the key to sort the highscores by. Defaults to
            the previous sort key.
        """
        if self._state.difficulty == "C":
            return
        if self._open_subwindows.get("highscores"):
            self._open_subwindows.get("highscores").close()
        if not settings:
            settings = HighscoreSettingsStruct(
                difficulty=get_difficulty(
                    self._state.x_size, self._state.y_size, self._state.mines
                ),
                per_cell=self._state.per_cell,
                drag_select=self._state.drag_select,
            )
        if sort_by:
            self._state.highscores_state.sort_by = sort_by
        self._state.highscores_state.name_hint = self._state.name
        win = highscores.HighscoresWindow(self, settings, self._state.highscores_state)
        win.show()
        self._open_subwindows["highscores"] = win

    def set_button_size(self, size: int) -> None:
        """Set the cell size in pixels."""
        self._state.btn_size = size
        self._mf_widget.update_btn_size(size)


class _CurrentInfoModal(QDialog):
    """A popup window with information about the current game."""

    def __init__(self, parent: QWidget, info: api.GameInfo, state_: state.State):
        super().__init__(parent)
        self._info = info
        self._state = state_
        self.setWindowTitle("Game info")
        self.setModal(True)
        self._setup_ui()

    def _setup_ui(self) -> None:
        """Set up the UI."""
        base_layout = QVBoxLayout(self)
        base_layout.addWidget(QLabel("Information about the current game:", self))
        info = self._info
        info_text = textwrap.dedent(
            f"""\
            {info.x_size} x {info.y_size} grid
            {info.mines} mines
            Max mines per cell: {info.per_cell}
            Drag-select: {self._state.drag_select}
            """
        )
        if info.finished_info:
            start_time = tm.strftime(
                "%Y-%m-%d %H:%M:%S", tm.localtime(info.finished_info.start_time)
            )
            fin_info = info.finished_info
            state = info.game_state.name.capitalize()
            info_text += textwrap.dedent(
                f"""\

                Game was started at {start_time}
                {state} after {fin_info.elapsed:.2f} seconds
                The board was {fin_info.prop_flagging * 100:.1f}% flagged
                The 3bv was: {fin_info.bbbv}
                The 3bv/s rate was: {fin_info.bbbvps:.2f}
                """
            )
            if info.game_state is GameState.LOST:
                info_text += textwrap.dedent(
                    f"""\

                    Remaining 3bv: {fin_info.rem_bbbv}
                    Game was {fin_info.prop_complete * 100:.1f}% complete
                    """
                )
                if fin_info.prop_complete > 0:
                    info_text += textwrap.dedent(
                        f"""\
                        Predicted completion time: {fin_info.elapsed / fin_info.prop_complete:.2f} seconds
                        """
                    )
        base_layout.addWidget(QLabel(info_text, self))
        ok_btn = QPushButton("Ok", self)
        ok_btn.pressed.connect(self.close)
        base_layout.addWidget(ok_btn)


class _CustomBoardModal(QDialog):
    """A popup window to select custom board dimensions."""

    def __init__(
        self, parent: QWidget, cols: int, rows: int, mines: int, callback: Callable
    ):
        super().__init__(parent)
        self._cols: int = cols
        self._rows: int = rows
        self._mines: int = mines
        self._callback: Callable = callback
        self.setWindowTitle("Custom")
        self.setModal(True)
        self._setup_ui()

    def _setup_ui(self) -> None:
        """Set up the UI."""
        base_layout = QVBoxLayout(self)
        base_layout.addWidget(
            QLabel("Select the number of rows, columns and mines.", self)
        )

        # Set up the input sliders and spin boxes.
        grid_layout = QGridLayout()
        base_layout.addLayout(grid_layout)

        def make_row(
            row: int, label: str, current: int, minval: int, maxval: int
        ) -> _SliderSpinner:
            grid_layout.addWidget(QLabel(label, self), row, 0)
            slider = _SliderSpinner(self)
            grid_layout.addWidget(slider, row, 1)
            slider.setValue(current)
            slider.setMinimum(minval)
            slider.setMaximum(maxval)
            return slider

        y_slider = make_row(1, "Rows", self._rows, 2, 50)
        x_slider = make_row(2, "Columns", self._cols, 2, 50)
        m_slider = make_row(3, "Mines", self._mines, 1, self._rows * self._cols - 1)

        def set_mines_max():
            m_slider.setMaximum(x_slider.value() * y_slider.value() - 1)

        x_slider.valueChanged.connect(set_mines_max)
        y_slider.valueChanged.connect(set_mines_max)

        # Ok/Cancel buttons.
        def ok_pressed():
            self._callback(x_slider.value(), y_slider.value(), m_slider.value())
            self.close()

        ok_btn = QPushButton("Ok", self)
        ok_btn.pressed.connect(ok_pressed)
        cancel_btn = QPushButton("Cancel", self)
        cancel_btn.pressed.connect(self.close)
        btns_layout = QHBoxLayout()
        base_layout.addLayout(btns_layout)
        btns_layout.addWidget(ok_btn)
        btns_layout.addWidget(cancel_btn)


class _ButtonSizeModal(QDialog):
    """A popup window to select the button size."""

    def __init__(self, parent: QWidget, btn_size: int, callback: Callable):
        super().__init__(parent)
        self.setWindowTitle("Button size")
        self._btn_size: int = btn_size
        self._callback: Callable = callback
        self.setModal(True)
        self._setup_ui()

    def _setup_ui(self) -> None:
        """Set up the UI."""
        base_layout = QVBoxLayout(self)
        base_layout.addWidget(QLabel("Select the button size in pixels.", self))

        # Set up the input sliders and spin boxes.
        slider = _SliderSpinner(self)
        slider.setValue(self._btn_size)
        slider.setMinimum(16)
        slider.setMaximum(40)
        base_layout.addWidget(slider)

        # Ok/Cancel buttons.
        def ok_pressed():
            self._callback(slider.value())
            self.close()

        ok_btn = QPushButton("Ok", self)
        ok_btn.pressed.connect(ok_pressed)
        cancel_btn = QPushButton("Cancel", self)
        cancel_btn.pressed.connect(self.close)
        btns_layout = QHBoxLayout()
        base_layout.addLayout(btns_layout)
        btns_layout.addWidget(ok_btn)
        btns_layout.addWidget(cancel_btn)


class _SliderSpinner(QWidget):
    """A combination of a slider and a spinbox."""

    valueChanged = pyqtSignal(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.slider = QSlider(Qt.Horizontal)
        self.numbox = QSpinBox()
        self.numbox.setRange(self.slider.minimum(), self.slider.maximum())
        self.numbox.setFixedWidth(50)
        self.slider.valueChanged.connect(self.numbox.setValue)
        self.slider.rangeChanged.connect(self.numbox.setRange)
        self.numbox.valueChanged.connect(self.slider.setValue)
        self.slider.valueChanged.connect(self.valueChanged.emit)
        self.valueChanged.connect(self.setValue)
        layout = QHBoxLayout(self)
        layout.addWidget(self.slider)
        layout.addWidget(self.numbox)

    def setMinimum(self, minval: int):
        self.slider.setMinimum(minval)

    def setMaximum(self, maxval: int):
        self.slider.setMaximum(maxval)

    def setValue(self, value: int):
        self.slider.setValue(value)

    def value(self) -> int:
        return self.slider.value()


class _NameEntryBar(QLineEdit):
    """Entry bar for entering a name."""

    name_updated_signal = pyqtSignal(str)

    def __init__(self, parent, text: str = ""):
        super().__init__(parent)
        self.setText(text)
        self.setPlaceholderText("Enter name here")
        self.setAlignment(Qt.AlignCenter)
        font = QFont("Helvetica")
        font.setBold(True)
        self.setFont(font)

    def keyPressEvent(self, event: QKeyEvent):
        super().keyPressEvent(event)
        if event.key() in [Qt.Key_Return, Qt.Key_Enter]:
            self.clearFocus()

    def focusOutEvent(self, event: QFocusEvent):
        super().focusOutEvent(event)
        self.setText(self.text().strip())
        logger.info("Setting name to %r", self.text())
        self.name_updated_signal.emit(self.text())


class _TextPopup(QWidget):
    """A popup window containing a block of text."""

    def __init__(self, parent: Optional[QWidget], title: str, file: os.PathLike):
        super().__init__(parent)
        self.setWindowFlag(Qt.Window)
        self.setWindowTitle(title)
        self.setMinimumSize(300, 300)

        self._text_widget = QLabel(self)
        self._ok_button = QPushButton(self)
        self._setup_ui()

        with open(file) as f:
            self._text_widget.setText(f.read())

    def _setup_ui(self) -> None:
        """Set up the UI layout."""
        lyt = QVBoxLayout(self)
        # lyt.setAlignment(Qt.AlignCenter)

        self._text_widget.setLineWidth(1)
        self._text_widget.setFrameShape(QFrame.Panel)
        self._text_widget.setMargin(10)
        self._text_widget.setStyleSheet(
            """
            background-color: white;
            """
        )
        self._text_widget.setWordWrap(True)
        lyt.addWidget(self._text_widget)

        self._ok_button.setText("Ok")
        self._ok_button.setMaximumWidth(50)
        self._ok_button.clicked.connect(self.close)
        lyt.addWidget(self._ok_button)


if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication
    import sys

    app = QApplication(sys.argv)
    widget = _CustomBoardModal(None, 10, 20, 30, lambda x, y, z: print(x, y, z))
    widget.show()
    app.exec()



from unittest import mock

import flask
import requests

import server
from server.bot import routes


app = flask.Flask(__name__)

app.add_url_rule("/bot/message", "bot_message", routes.bot_message, methods=["POST"])


def send_bot_message():
    url = "http://localhost:5000/bot"
    data = {
        # Webhook info
        "id": "Y2lzY29zcGFyazovL3VzL1dFQkhPT0svZjRlNjA1NjAtNjYwMi00ZmIwLWEyNWEtOTQ5ODgxNjA5NDk3",
        "name": "New message in 'Project Unicorn' room",
        "resource": "messages",
        "event": "created",
        "filter": "roomId=Y2lzY29zcGFyazovL3VzL1JPT00vYmJjZWIxYWQtNDNmMS0zYjU4LTkxNDctZjE0YmIwYzRkMTU0",
        "orgId": "OTZhYmMyYWEtM2RjYy0xMWU1LWExNTItZmUzNDgxOWNkYzlh",
        "createdBy": "Y2lzY29zcGFyazovL3VzL1BFT1BMRS9mNWIzNjE4Ny1jOGRkLTQ3MjctOGIyZi1mOWM0NDdmMjkwNDY",
        "appId": "Y2lzY29zcGFyazovL3VzL0FQUExJQ0FUSU9OL0MyNzljYjMwYzAyOTE4MGJiNGJkYWViYjA2MWI3OTY1Y2RhMzliNjAyOTdjODUwM2YyNjZhYmY2NmM5OTllYzFm",
        "ownedBy": "creator",
        "status": "active",
        "actorId": "Y2lzY29zcGFyazovL3VzL1BFT1BMRS9mNWIzNjE4Ny1jOGRkLTQ3MjctOGIyZi1mOWM0NDdmMjkwNDY",
        # Message data
        "data": {
            "id": "Y2lzY29zcGFyazovL3VzL01FU1NBR0UvOTJkYjNiZTAtNDNiZC0xMWU2LThhZTktZGQ1YjNkZmM1NjVk",
            "roomId": "Y2lzY29zcGFyazovL3VzL1JPT00vYmJjZWIxYWQtNDNmMS0zYjU4LTkxNDctZjE0YmIwYzRkMTU0",
            "personId": "Y2lzY29zcGFyazovL3VzL1BFT1BMRS9mNWIzNjE4Ny1jOGRkLTQ3MjctOGIyZi1mOWM0NDdmMjkwNDY",
            "personEmail": "matt@example.com",
            "created": "2015-10-18T14:26:16.000Z",
        },
    }
    requests.post(url, json=data)


def receive_bot_message(msg: str):
    post_data = {
        "data": {
            "id": "foo",
            "personEmail": "person@foo.com",
            "roomId": "room",
            "personId": "person",
        }
    }
    resp = mock.Mock()
    resp.json.return_value = {
        "id": "msg ID",
        "roomId": "room ID",
        "roomType": "group",
        "toPersonId": "to-person ID",
        "toPersonEmail": "julie@example.com",
        "text": msg,
        "markdown": msg,
        "personId": "person ID",
        "personEmail": "matt@example.com",
        "mentionedPeople": ["person ID 1", "person ID 2"],
        "mentionedGroups": ["all"],
        "attachments": [],
        "files": [],
        "created": "2015-10-18T14:26:16+00:00",
    }
    with server.utils.multiple_contexts(
        app.test_client(), mock.patch("requests.get", return_value=resp)
    ) as ctx:
        tc, _ = ctx
        return tc.post("/bot/message", json=post_data)



"""
types.py - Type definitions

June 2018, Lewis Gaul

Exports:
CellContentsType (class)
    Base class for cell contents types.
CellMineType (class)
    Base class for cell contents of a mine type.
CellUnclicked, CellNum, CellMine, CellHitMine, CellFlag, CellwrongFlag (class)
    CellContentsType implementations.

GameState (Enum)
    The possible states of a game.

GameFlagMode (Enum)
    The possible flagging modes for a game.
"""

import enum
import functools


# ------------------------------------------------------------------------------
# Cell contents types
# ------------------------------------------------------------------------------


class _NumericCellContentsMixin:
    """
    A mixin for numeric cell contents types, allowing adding and subtracting integers.
    """

    char: str

    def __init__(self, num):
        if not isinstance(num, int):
            raise TypeError("Number should be an integer")
        self.num = num

    def __repr__(self):
        return self.char + str(self.num)

    def __add__(self, obj):
        if type(obj) is not int:
            raise TypeError("Can only add integers to cell contents types")
        else:
            return self.__class__(self.num + obj)

    def __sub__(self, obj):
        if type(obj) is not int:
            raise TypeError("Can only subtract integers from cell contents types")
        else:
            return self.__class__(self.num - obj)


class CellContentsType:
    """Abstract base class for contents of a minesweeper board cell."""

    char: str

    @functools.lru_cache(maxsize=None)
    def __new__(cls, *args):
        if cls == CellContentsType:
            raise TypeError("Base class should not be instantiated")
        return super().__new__(cls)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return self.char


class CellUnclicked(CellContentsType):
    """Unclicked cell on a minesweeper board."""

    char = "#"


class CellNum(_NumericCellContentsMixin, CellContentsType):
    """Number shown in a cell on a minesweeper board."""

    char = ""

    def __init__(self, num):
        super().__init__(num)
        if num < 0:
            raise ValueError("Cell value cannot be negative")


class CellMineType(_NumericCellContentsMixin, CellContentsType):
    """Abstract base class for the number of a mine type in a cell."""

    def __new__(cls, num=1):
        if cls == CellMineType:
            raise TypeError(
                f"{type(cls)} should be used as a base class and not "
                "instantiated directly"
            )
        return super().__new__(cls, num)

    def __init__(self, num=1):
        super().__init__(num)
        if num < 1:
            raise ValueError("Mine-type cell contents must represent one or more mines")

    @staticmethod
    def get_class_from_char(char):
        """
        Get the class of mine-like cell contents using the character
        representation.

        Arguments:
        char (str, length 1)
            The character representation of a cell contents type.

        Return:
            The cell contents class.
        """
        for cls in [CellMine, CellHitMine, CellFlag, CellWrongFlag]:
            if cls.char == char:
                return cls


class CellMine(CellMineType):
    """Number of mines in a cell shown on a minesweeper board."""

    char = "M"


class CellHitMine(CellMineType):
    """Number of hit mines in a cell shown on a minesweeper board."""

    char = "!"


class CellFlag(CellMineType):
    """Number of flags in a cell shown on a minesweeper board."""

    char = "F"


class CellWrongFlag(CellFlag):
    """Number of incorrect flags in a cell shown on a minesweeper board."""

    char = "X"


# ------------------------------------------------------------------------------
# Game enums
# ------------------------------------------------------------------------------


class GameState(str, enum.Enum):
    """
    Enum representing the state of a game.
    """

    READY = "READY"
    ACTIVE = "ACTIVE"
    WON = "WON"
    LOST = "LOST"

    def started(self) -> bool:
        return self is not self.READY

    def finished(self) -> bool:
        return self in [self.WON, self.LOST]


# ------------------------------------------------------------------------------
# GUI enums
# ------------------------------------------------------------------------------


class FaceState(enum.Enum):
    READY = "ready"
    ACTIVE = "active"
    WON = "won"
    LOST = "lost"


class CellImageType(enum.Flag):
    BUTTONS = enum.auto()
    NUMBERS = enum.auto()
    MARKERS = enum.auto()
    ALL = BUTTONS | NUMBERS | MARKERS


class UIMode(enum.Enum):
    GAME = enum.auto()
    CREATE = enum.auto()



__all__ = ("is_highscore_new_best", "multiple_contexts")

import contextlib
from typing import Optional

from minegauler.shared import highscores as hs


def is_highscore_new_best(h: hs.HighscoreStruct) -> Optional[str]:
    all_highscores = hs.get_highscores(hs.HighscoresDatabases.REMOTE, settings=h)
    return hs.is_highscore_new_best(h, all_highscores)


@contextlib.contextmanager
def multiple_contexts(*contexts):
    """
    Context manager to activate multiple context managers.

    :param contexts:
        The context managers to activate.
    """
    stack = contextlib.ExitStack()
    entered = []
    for ctx in contexts:
        entered.append(stack.enter_context(ctx))
    try:
        yield tuple(entered)
    finally:
        stack.close()



"""
msgparse.py - Parse bot messages

February 2020, Lewis Gaul
"""

__all__ = ("RoomType", "parse_msg")

import argparse
import enum
import logging
import sys
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from minegauler.shared import highscores as hs

from . import formatter, utils


logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# Message parsing
# ------------------------------------------------------------------------------


class InvalidArgsError(Exception):
    pass


class PositionalArg:
    def __init__(
        self,
        name: str,
        *,
        parse_name: bool = False,
        nargs: Union[int, str] = 1,
        default: Any = None,
        choices: Optional[Iterable[Any]] = None,
        type: Optional[Callable] = None,
        validate: Optional[Callable] = None,
    ):
        if (isinstance(nargs, int) and nargs < 1) and nargs not in ["?", "*", "+"]:
            raise ValueError(f"Bad nargs value {nargs!r}")
        self.name = name
        self.parse_name = parse_name
        self.nargs = nargs
        if default is None and nargs not in [1, "?"]:
            self.default = []
        else:
            self.default = default
        self._choices = choices
        self._type = type
        self._validate = validate

    def convert(self, value):
        if self._type is not None:
            return self._type(value)
        return value

    def validate(self, value):
        if self._choices is not None and value not in self._choices:
            return False
        if self._validate:
            return self._validate(value)
        return True


class ArgParser(argparse.ArgumentParser):
    """
    A specialised arg parser.

    The list of args to be parsed can contain the following, with no overlap:
     - Ordered positional args at the start
     - Regular argparse args with no order

    Positional args are added with 'add_positional_arg()'. The order of calls to
    this method determines the order the args must appear in. The following
    options are accepted:
     - parse_name: Whether the arg name should be parsed.
     - nargs: The number of args to accept. Accepted values are positive
        integers, "?" for 0 or 1, "*" for 0 or more, "+" for 1 or more.
     - choices: As for 'add_argument()'.
     - type: As for 'add_argument()'.

    Positional args are greedily consumed, but if an arg does not satisfy
    choices or if the 'type' callable raises an exception then no more values
    will be accepted for that arg.

    Examples:
     - If a positional arg has no type/choices restrictions and unbounded
        'nargs' (i.e. set to "*" or "+") then all positional options will be
        consumed by this arg.
     - If a positional arg has unbounded 'nargs' but has a 'type' that raises an
        exception for invalid args, subsequent args will take over the matching.
        Note that if insufficient options are matched (e.g. nargs="+" and no
        options are matched) then parsing ends with a standard error.

    Positional argument parsing ends as soon as an option starting with a dash
    is encountered, or when the positional args are exhausted. The remaining
    options are passed to a standard argparse parser to match args added with
    'add_argument()'.
    """

    def __init__(self):
        super().__init__(add_help=False)
        self._name_parse_args = []
        self._positional_args = []

    def parse_known_args(self, args: Iterable[str], namespace=None):
        """
        Override the default behaviour.
        """
        if namespace is None:
            namespace = argparse.Namespace()

        # Start with positional args.
        args = self._parse_positional_args(args, namespace)

        # Replace optional args so that they don't require the dashes.
        for i, arg in enumerate(args):
            if arg in self._name_parse_args:
                args[i] = f"--{arg}"

        # Regular parsing of remaining args
        return super().parse_known_args(args, namespace)

    def add_argument(self, name, *args, **kwargs):
        """
        Override the default behaviour.
        """
        if not name.startswith("--"):
            self._name_parse_args.append(name)
            name = "--" + name.lstrip("-")
        super().add_argument(name, *args, **kwargs)

    def error(self, message):
        raise InvalidArgsError(message)

    def add_positional_arg(self, name: str, **kwargs) -> None:
        """
        Add a positional argument for parsing.

        :param name:
            The name of the argument - must be unique.
        :param kwargs:
            Arguments to pass to the 'PositionalArg' class.
        """
        self._positional_args.append(PositionalArg(name, **kwargs))

    def _parse_positional_args(self, kws: Iterable[str], namespace) -> Iterable[str]:
        """
        Parse the positional args.

        :param kws:
            The provided keywords.
        :param namespace:
            The namespace to set argument values in.
        :return:
            The remaining unmatched keywords.
        """
        for arg in self._positional_args:
            result, kws = self._parse_single_positional_arg(arg, kws)
            setattr(namespace, arg.name, result)
        return kws

    def _parse_single_positional_arg(
        self, arg: PositionalArg, kws: Iterable[str]
    ) -> Tuple[Any, Iterable[str]]:
        """
        Parse a single positional arg. Raise InvalidArgsError if not enough
        matching args are found.
        """
        required = arg.nargs not in ["?", "*"]
        if isinstance(arg.nargs, int):
            max_matches = arg.nargs
            exp_args_string = str(arg.nargs)
        elif arg.nargs == "?":
            max_matches = 1
            exp_args_string = "optionally one"
        elif arg.nargs == "*":
            max_matches = None
            exp_args_string = "any number of"
        elif arg.nargs == "+":
            max_matches = None
            exp_args_string = "at least one"
        else:
            assert False

        # First parse the arg name if required.
        if kws and arg.parse_name:
            if kws[0] == arg.name:  # Found arg
                kws.pop(0)
            elif not required:  # No match
                return arg.default, kws
            else:
                raise InvalidArgsError(f"Expected to find {arg.name!r}")

        # Now parse argument values.
        matches = []
        while kws and (max_matches is None or len(matches) < max_matches):
            try:
                kw_value = arg.convert(kws[0])
                assert arg.validate(kw_value)
            except Exception as e:
                logger.debug(e)
                if arg.parse_name and not matches:
                    # We parsed the name of the arg, so we expected to find
                    # at least one value...
                    raise InvalidArgsError(
                        f"Got name of positional arg {arg.name!r} but no values"
                    )
                else:
                    break
            else:
                matches.append(kw_value)
                kws.pop(0)

        if required and not matches:
            raise InvalidArgsError(f"Expected {exp_args_string} {arg.name!r} arg")
        elif isinstance(arg.nargs, int) and len(matches) != arg.nargs:
            assert len(matches) < arg.nargs
            raise InvalidArgsError(f"Expected {exp_args_string} {arg.name!r} arg")

        arg_value = arg.default
        if matches:
            if arg.nargs in [1, "?"]:
                assert len(matches) == 1
                arg_value = matches[0]
            else:
                arg_value = matches

        return arg_value, kws


class BotMsgParser(ArgParser):
    def add_username_arg(self, *, nargs: Union[int, str] = 1, allow_me=True):
        choices = list(utils.USER_NAMES)
        if allow_me:
            choices.append("me")
        self.add_positional_arg("username", nargs=nargs, choices=choices)

    def add_difficulty_arg(self):
        self.add_positional_arg(
            "difficulty", nargs="?", type=self._convert_difficulty_arg
        )

    def add_rank_type_arg(self):
        def convert(arg):
            try:
                return self._convert_difficulty_arg(arg)
            except InvalidArgsError:
                raise  # TODO
                # if arg == "combined":
                #     return "combined"
                # elif arg == "official":
                #     return "official"
                # else:
                #     raise InvalidArgsError(f"Invalid rank type {arg!r}")

        self.add_positional_arg("rank_type", nargs="?", type=convert)

    def add_per_cell_arg(self):
        self.add_argument("per-cell", type=int, choices=[1, 2, 3])

    def add_drag_select_arg(self):
        def _arg_type(arg):
            if arg == "on":
                return True
            elif arg == "off":
                return False
            else:
                raise InvalidArgsError("Drag select should be one of {'on', 'off'}")

        self.add_argument("drag-select", type=_arg_type)

    @staticmethod
    def _convert_difficulty_arg(arg):
        if arg in ["b", "beginner"]:
            return "beginner"
        elif arg in ["i", "intermediate"]:
            return "intermediate"
        elif arg in ["e", "expert"]:
            return "expert"
        elif arg in ["m", "master"]:
            return "master"
        else:
            raise InvalidArgsError(f"Invalid difficulty {arg!r}")


# ------------------------------------------------------------------------------
# Message handling
# ------------------------------------------------------------------------------


CommandMapping = Dict[Optional[str], Union[Callable, "CommandMapping"]]


class RoomType(enum.Enum):
    GROUP = "group"
    DIRECT = "direct"

    def to_cmds(self) -> CommandMapping:
        if self is self.GROUP:
            return _GROUP_COMMANDS
        else:
            assert self is self.DIRECT
            return _DIRECT_COMMANDS


_GENERAL_INFO = """\
Welcome to the Minegauler bot!

Download Minegauler from https://github.com/LewisGaul/minegauler.

The bot provides the ability to check highscores and matchups for games of \
Minegauler, and also gives notifications whenever anyone in the group sets a \
new highscore.

There are a few settings to filter Minegauler games with:
 - difficulty: One of 'beginner', 'intermediate', 'expert' or 'master'.  
   By default the combination of beginner, intermediate and expert is used, \
   with 1000 used for any difficulties not completed.
 - drag-select: One of 'on' or 'off'.  
   By default no filter is applied (the best time of either mode is used).
 - per-cell: One of 1, 2 or 3.  
   By default no filter is applied (the best time of any mode is used).

All highscores are independent of each of the above modes, and all commands \
accept these filters to view specific times. E.g. 'ranks beginner per-cell 1'.

Commands can be issued in group chat and direct chat. Most of the commands are \
the same, however some are only available in one of the other (e.g. \
'set nickname'). Type 'help' or '?' to check the available commands.

To interact with a bot in Webex Teams group chats, the bot must be tagged. To \
do this, type @ followed by the bot's name. The client should auto-suggest a \
completion, at which point you'll need to press enter or tab to select the \
completion and turn it into a tag.

Some useful common commands:  
'ranks' - display rankings  
'matchups <name> <name> ...' - display comparison of times between users

Some useful group-chat commands:  
'challenge <name> ...' - challenge other users to a game

Some useful direct-chat commands:  
'set nickname' - set your nickname that you use in the Minegauler app
"""


def helpstring(text):
    def decorator(func):
        func.__helpstring__ = text
        return func

    return decorator


def schema(text):
    def decorator(func):
        func.__schema__ = text
        return func

    return decorator


def cmd_help(
    func: Callable, *, only_schema: bool = False, allow_markdown: bool = False
) -> str:
    lines = []
    if not only_schema:
        try:
            lines.append(func.__helpstring__)
        except AttributeError:
            logger.warning(
                "No helpstring found on message handling function %r", func.__name__
            )
    try:
        schema = func.__schema__
    except AttributeError:
        logger.warning("No schema found on message handling function %r", func.__name__)
    else:
        if allow_markdown:
            lines.append(f"`{schema}`")
        else:
            lines.append(schema)

    if not lines:
        return "Unexpected error: unable to get help message\n\n"

    return "\n\n".join(lines)


@helpstring("Get help for a command")
@schema("help [<command>]")
def group_help(args, **kwargs):
    allow_markdown = kwargs.get("allow_markdown", False)

    linebreak = "\n\n" if allow_markdown else "\n"
    commands = _flatten_cmds(_GROUP_COMMANDS)
    if allow_markdown:
        commands = f"`{commands}`"

    if not args:
        return commands

    func, _ = _map_to_cmd(" ".join(args), _GROUP_COMMANDS)
    if func is None:
        raise InvalidArgsError(linebreak.join(["Unrecognised command", commands]))
    else:
        return cmd_help(func, allow_markdown=allow_markdown)


@helpstring("Get help for a command")
@schema("help [<command>]")
def direct_help(args, **kwargs):
    allow_markdown = kwargs.get("allow_markdown", False)

    linebreak = "\n\n" if allow_markdown else "\n"
    commands = _flatten_cmds(_DIRECT_COMMANDS)
    if allow_markdown:
        commands = f"`{commands}`"

    if not args:
        return commands

    func, _ = _map_to_cmd(" ".join(args), _DIRECT_COMMANDS)
    if func is None:
        raise InvalidArgsError(linebreak.join(["Unrecognised command", commands]))
    else:
        return cmd_help(func, allow_markdown=allow_markdown)


@helpstring("Get information about the game")
@schema("info")
def info(args, **kwargs):
    # Check no args given.
    BotMsgParser().parse_args(args)
    return _GENERAL_INFO


@helpstring("Get player info")
@schema(
    "player <name> [b[eginner] | i[ntermediate] | e[xpert] | m[aster]] "
    "[drag-select {on | off}] [per-cell {1 | 2 | 3}]"
)
def player(args, username: str, allow_markdown=False, **kwargs):
    parser = BotMsgParser()
    parser.add_username_arg()
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    if args.username == "me":
        args.username = username

    diff = args.difficulty[0] if args.difficulty else None
    highscores = hs.get_highscores(
        hs.HighscoresDatabases.REMOTE,
        name=utils.USER_NAMES[args.username],
        difficulty=diff,
        drag_select=args.drag_select,
        per_cell=args.per_cell,
    )

    filters_str = formatter.format_filters(
        None, args.drag_select, args.per_cell, no_difficulty=True
    )
    if filters_str:
        filters_str = " with " + filters_str
    lines = ["Player info for {}{}:".format(args.username, filters_str)]
    lines.extend(
        formatter.format_player_highscores(highscores, difficulty=args.difficulty)
    )

    linebreak = "  \n" if allow_markdown else "\n"

    return linebreak.join(lines)


@helpstring("Get rankings")
@schema(
    "ranks [b[eginner] | i[ntermediate] | e[xpert] | m[aster]] "
    "[drag-select {on | off}] [per-cell {1 | 2 | 3}]"
)
def ranks(args, **kwargs) -> str:
    allow_markdown = kwargs.get("allow_markdown", False)

    parser = BotMsgParser()
    parser.add_rank_type_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)

    times = utils.get_highscore_times(args.rank_type, args.drag_select, args.per_cell)

    lines = []
    lines.append(
        "Rankings for {}".format(
            formatter.format_filters(args.rank_type, args.drag_select, args.per_cell)
        )
    )
    ranks = formatter.format_highscore_times(times)
    if allow_markdown:
        ranks = f"```\n{ranks}\n```"
    lines.append(ranks)

    return "\n".join(lines)


@helpstring("Get stats for played games")
@schema(
    "stats [players ...] [b[eginner] | i[ntermediate] | e[xpert] | m[aster]] "
    "[drag-select {on | off}] [per-cell {1 | 2 | 3}]"
)
def stats(args, **kwargs):
    parser = BotMsgParser()
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    return "Stats"


@helpstring("Get player stats")
@schema(
    "stats players {all | <name> [<name> ...]} "
    "[b[eginner] | i[ntermediate] | e[xpert] | m[aster]] "
    "[drag-select {on | off}] [per-cell {1 | 2 | 3}]"
)
def stats_players(args, **kwargs):
    parser = BotMsgParser()
    parser.add_username_arg(nargs="+")
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    return "Player stats {}".format(", ".join(args.username))


@helpstring("Get matchups for given players")
@schema(
    "matchups <name> <name> [<name> ...] "
    "[b[eginner] | i[ntermediate] | e[xpert] | m[aster]] "
    "[drag-select {on | off}] [per-cell {1 | 2 | 3}]"
)
def matchups(
    args,
    username: str,
    allow_markdown: bool = False,
    room_type=RoomType.DIRECT,
    **kwargs,
):
    parser = BotMsgParser()
    parser.add_username_arg(nargs="+")
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    users = {u if u != "me" else username for u in args.username}
    if len(users) < 2 or len(users) > 5:
        raise InvalidArgsError("Require between 2 and 5 username args")
    names = {utils.USER_NAMES[u] for u in users}

    times = utils.get_highscore_times(
        args.difficulty, args.drag_select, args.per_cell, names
    )
    matchups = utils.get_matchups(times)

    if allow_markdown and room_type is RoomType.GROUP:
        users_str = ", ".join(utils.tag_user(u) for u in users)
    else:
        users_str = ", ".join(users)
    lines = [
        "Matchups between {} for {}:".format(
            users_str,
            formatter.format_filters(args.difficulty, args.drag_select, args.per_cell),
        )
    ]
    lines.extend(formatter.format_matchups(matchups))
    linebreak = "  \n" if allow_markdown else "\n"

    return linebreak.join(lines)


@helpstring("Get best matchups including at least one of specified players")
@schema(
    "best-matchups [<name> ...] "
    "[b[eginner] | i[ntermediate] | e[xpert] | m[aster]] "
    "[drag-select {on | off}] [per-cell {1 | 2 | 3}]"
)
def best_matchups(
    args, username: str, allow_markdown=False, room_type=RoomType.DIRECT, **kwargs
):
    parser = BotMsgParser()
    parser.add_username_arg(nargs="*")
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    users = {u if u != "me" else username for u in args.username}
    names = {utils.USER_NAMES[u] for u in users}

    times = utils.get_highscore_times(
        args.difficulty, args.drag_select, args.per_cell, utils.USER_NAMES.values()
    )
    matchups = utils.get_matchups(times, include_users=names)[:10]

    if allow_markdown and room_type is RoomType.GROUP:
        users_str = ", ".join(utils.tag_user(u) for u in users)
    else:
        users_str = ", ".join(users)
    if users_str:
        users_str = " including " + users_str
    lines = [
        "Best matchups{} for {}:".format(
            users_str,
            formatter.format_filters(args.difficulty, args.drag_select, args.per_cell),
        )
    ]
    lines.extend(formatter.format_matchups(matchups))
    linebreak = "  \n" if allow_markdown else "\n"

    return linebreak.join(lines)


@helpstring("Challenge other players to a game")
@schema(
    "challenge <name> [<name> ...] "
    "[b[eginner] | i[ntermediate] | e[xpert] | m[aster]] "
    "[drag-select {on | off}] [per-cell {1 | 2 | 3}]"
)
def challenge(args, username: str, **kwargs):
    parser = BotMsgParser()
    parser.add_username_arg(nargs="+", allow_me=False)
    parser.add_difficulty_arg()
    parser.add_per_cell_arg()
    parser.add_drag_select_arg()
    args = parser.parse_args(args)
    names = {u for u in args.username if u != username}
    if len(names) < 1:
        raise InvalidArgsError("Need at least one other player's name")

    users_str = ", ".join(utils.tag_user(u) for u in names)
    diff_str = args.difficulty + " " if args.difficulty else ""
    opts = dict()
    if args.drag_select:
        opts["drag-select"] = "on" if args.drag_select else "off"
    if args.per_cell:
        opts["per-cell"] = args.per_cell
    if opts:
        opts_str = " with {}".format(formatter.format_kwargs(opts))
    else:
        opts_str = ""

    return "{} has challenged {} to a {}game of Minegauler{}".format(
        username, users_str, diff_str, opts_str
    )


@helpstring("Set your nickname")
@schema("set nickname <name>")
def set_nickname(args, username: str, **kwargs):
    new = " ".join(args)
    old = utils.USER_NAMES[username]
    logger.debug("Changing nickname of %s from %r to %r", username, old, new)
    utils.set_user_nickname(username, new)
    return f"Nickname changed from '{old}' to '{new}'"


# fmt: off
_COMMON_COMMANDS = {
    "help": info,
    "info": info,
    "player": player,
    "ranks": ranks,
    # "stats": {
    #     None: stats,
    #     "players": stats_players,
    # },
    "matchups": matchups,
    "best-matchups": best_matchups,
}

_GROUP_COMMANDS = {
    **_COMMON_COMMANDS,
    "help": group_help,
    "challenge": challenge,
}

_DIRECT_COMMANDS = {
    **_COMMON_COMMANDS,
    "help": direct_help,
    "set": {
        "nickname": set_nickname,
    },
}
# fmt: on


def _map_to_cmd(msg: str, cmds: CommandMapping) -> Tuple[Callable, List[str]]:
    func = None
    words = msg.split()

    while words:
        next_word = words[0]
        if next_word in cmds:
            words.pop(0)
            if callable(cmds[next_word]):
                func = cmds[next_word]
                break
            else:
                cmds = cmds[next_word]
        else:
            break
        if None in cmds:
            func = cmds[None]

    return func, words


def _flatten_cmds(cmds: CommandMapping, root: bool = True) -> str:
    to_join = []
    for k, v in cmds.items():
        item = k
        if isinstance(v, dict):
            item += " " + _flatten_cmds(v, root=False)
        if item:
            to_join.append(item)
    ret = " | ".join(to_join)
    if None in cmds:
        ret = f"[{ret}]"
    elif len(cmds) > 1 and not root:
        ret = f"{{{ret}}}"
    return ret


def parse_msg(
    msg: str, room_type: RoomType, *, allow_markdown: bool = False, **kwargs
) -> str:
    """
    Parse a message and perform the corresponding action.

    :param msg:
        The message to parse.
    :param room_type:
        The room type the message was received in.
    :param allow_markdown:
        Whether to allow markdown in the response.
    :param kwargs:
        Other arguments to pass on to sub-command functions.
    :return:
        Response text.
    :raise InvalidArgsError:
        Unrecognised command. The text of the error is a suitable error/help
        message.
    """
    msg = msg.strip()
    if msg.endswith("?") and not (msg.startswith("help ") and msg.split()[1] != "?"):
        # Change "?" at the end to be treated as help, except for the case where
        # this is a double help intended for a subcommand, e.g.
        # 'help matchups ?'.
        msg = "help " + msg[:-1]

    cmds = room_type.to_cmds()
    func = None
    try:
        func, args = _map_to_cmd(msg, cmds)
        if func is None:
            raise InvalidArgsError("Base command not found")
        return func(args, allow_markdown=allow_markdown, room_type=room_type, **kwargs)
    except InvalidArgsError as e:
        logger.debug("Invalid message: %r", msg)
        if func is None:
            return "Unrecognised command - try 'help' or 'info'"
        else:
            linebreak = "\n\n" if allow_markdown else "\n"
            resp_msg = cmd_help(func, only_schema=True, allow_markdown=allow_markdown)
            resp_msg = linebreak.join(["Unrecognised command", resp_msg])

        raise InvalidArgsError(resp_msg) from e


# ------------------------------------------------------------------------------
# Main
# ------------------------------------------------------------------------------


def main(argv):
    try:
        resp = parse_msg(" ".join(argv), RoomType.GROUP, username="dummy-user")
    except InvalidArgsError as e:
        resp = str(e)
    print(resp)


if __name__ == "__main__":
    main(sys.argv[1:])



__all__ = ("add_new_highscore_hook", "get_new_highscore_hooks", "utils")

from typing import Callable, Iterable

from minegauler.shared import highscores as hs

from . import utils


_new_highscore_hooks = []


def add_new_highscore_hook(func: Callable[[hs.HighscoreStruct], None]):
    _new_highscore_hooks.append(func)


def get_new_highscore_hooks() -> Iterable[Callable[[hs.HighscoreStruct], None]]:
    return iter(_new_highscore_hooks)



"""
__init__.py - Available imports from the package

December 2018, Lewis Gaul
"""

__all__ = ("IN_EXE", "ROOT_DIR", "SETTINGS_FILE")

import pathlib
import sys


IN_EXE: bool = hasattr(sys, "frozen") and hasattr(sys, "_MEIPASS")

ROOT_DIR: pathlib.Path
if IN_EXE:
    ROOT_DIR = pathlib.Path(__file__).parent.parent
else:
    ROOT_DIR = pathlib.Path(__file__).parent

SETTINGS_FILE: pathlib.Path = ROOT_DIR / "settings.cfg"



"""
utils.py - General utilities

December 2018, Lewis Gaul
"""

__all__ = ("StructConstructorMixin",)

import logging
from typing import Any, Dict

import attr


logger = logging.getLogger(__name__)


class StructConstructorMixin:
    """
    A mixin class adding methods for ways to create instances.
    """

    @classmethod
    def from_structs(cls, *structs):
        """
        Create an instance using namespace(s) containing the required fields.

        Later arguments take precedence.
        """
        dict_ = {}
        for struct in structs:
            dict_.update(attr.asdict(struct))
        return cls.from_dict(dict_)

    @classmethod
    def from_dict(cls, dict_: Dict[str, Any]):
        """
        Create an instance from a dictionary.

        Ignores extra attributes.
        """
        args = {a: v for a, v in dict_.items() if a in attr.fields_dict(cls)}
        return cls(**args)

    def copy(self):
        """
        Create and return a copy of the instance.

        This is a shallow copy.
        """
        return self.from_structs(self)



"""
engine.py - The core game logic

November 2018, Lewis Gaul

Exports:
BaseController (class)
    Implementation of game logic and provision of functions to be called by a
    frontend implementation.
"""

__all__ = ("BaseController", "CreateController", "GameController")

import json
import logging
import os
from typing import Dict, Optional

import attr

from ..shared import utils
from ..types import (
    CellContentsType,
    CellFlag,
    CellMine,
    CellNum,
    CellUnclicked,
    GameState,
    UIMode,
)
from ..typing import Coord_T
from . import api
from . import board as brd
from . import game


logger = logging.getLogger(__name__)


def _save_minefield(mf: brd.Minefield, file: os.PathLike) -> None:
    """
    Save a minefield to file.

    :param mf:
        The minefield to save.
    """
    if os.path.isfile(file):
        logger.warning("Overwriting file at %s", file)
    elif os.path.isdir(file):
        logger.error("Unable to save minefield - directory exists at %s", file)
        return
    with open(file, "w") as f:
        json.dump(mf.to_json(), f)


@attr.attrs(auto_attribs=True)
class _SharedInfo:
    """
    Information to pass to frontends.
    
    Elements:
    cell_updates ({(int, int): CellContentsType, ...})
        Dictionary of updates to cells, mapping the coordinate to the new
        contents of the cell.
    game_state (GameState)
        The state of the game.
    mines_remaining (int)
        The number of mines remaining to be found, given by
        [total mines] - [number of flags]. Can be negative if there are too many
        flags. If the number is unchanged, None may be passed.
    lives_remaining (int)
        The number of lives remaining.
    finish_time (float | None)
        The time elapsed if the game has ended, otherwise None.
    """

    cell_updates: Dict[Coord_T, CellContentsType] = attr.Factory(dict)
    game_state: GameState = GameState.READY
    mines_remaining: int = 0
    lives_remaining: int = 0
    end_game_info: Optional[api.EndedGameInfo] = None


class BaseController(api.AbstractSwitchingController):
    """Base controller implementing all user interaction methods."""

    def __init__(
        self, opts: utils.GameOptsStruct, *, notif: Optional[api.Caller] = None
    ):
        super().__init__(opts, notif=notif)
        self._mode = UIMode.GAME
        self._active_ctrlr: api.AbstractController = GameController(
            self._opts, notif=self._notif
        )

    def reinitialise(self):
        """Reinitialise the concrete game controller."""
        self._mode = UIMode.GAME
        self._active_ctrlr: api.AbstractController = GameController(
            self._opts, notif=self._notif
        )

    def switch_mode(self, mode: UIMode, *args, **kwargs) -> None:
        """Switch the mode of the UI, e.g. into 'create' mode."""
        super().switch_mode(mode)
        if mode is self._mode:
            logger.debug("Ignore switch mode request because mode is already %s", mode)
            return
        if mode is UIMode.GAME:
            self._active_ctrlr = GameController(
                self._opts, notif=self._notif, *args, **kwargs
            )
        elif mode is UIMode.CREATE:
            self._active_ctrlr = CreateController(
                self._opts, notif=self._notif, *args, **kwargs
            )
        else:
            raise ValueError(f"Unrecognised UI mode: {mode}")
        self._mode = mode

    # ----------------------------------
    # Delegated abstractmethods
    # ----------------------------------
    @property
    def board(self) -> brd.Board:
        return self._active_ctrlr.board

    def get_game_info(self) -> api.GameInfo:
        return self._active_ctrlr.get_game_info()

    def new_game(self) -> None:
        self._active_ctrlr.new_game()

    def restart_game(self) -> None:
        self._active_ctrlr.restart_game()

    def select_cell(self, coord: Coord_T) -> None:
        self._active_ctrlr.select_cell(coord)

    def flag_cell(self, coord: Coord_T, *, flag_only: bool = False) -> None:
        self._active_ctrlr.flag_cell(coord, flag_only=flag_only)

    def chord_on_cell(self, coord: Coord_T) -> None:
        self._active_ctrlr.chord_on_cell(coord)

    def remove_cell_flags(self, coord: Coord_T) -> None:
        self._active_ctrlr.remove_cell_flags(coord)

    def resize_board(self, x_size: int, y_size: int, mines: int) -> None:
        self._opts.x_size = x_size
        self._opts.y_size = y_size
        self._opts.mines = mines
        self._active_ctrlr.resize_board(x_size=x_size, y_size=y_size, mines=mines)

    def set_first_success(self, value: bool) -> None:
        self._opts.first_success = value
        self._active_ctrlr.set_first_success(value)

    def set_per_cell(self, value: int) -> None:
        self._opts.per_cell = value
        self._active_ctrlr.set_per_cell(value)

    def save_current_minefield(self, file: os.PathLike) -> None:
        self._active_ctrlr.save_current_minefield(file)

    def load_minefield(self, file: os.PathLike) -> None:
        with open(file) as f:
            mf = brd.Minefield.from_json(json.load(f))
        self._opts.x_size = mf.x_size
        self._opts.y_size = mf.y_size
        self._opts.mines = mf.nr_mines

        if self._mode is UIMode.CREATE:
            self.switch_mode(UIMode.GAME, minefield_file=file)
            self._notif.switch_mode(UIMode.GAME)
        else:
            self._active_ctrlr.load_minefield(file)


class GameController(api.AbstractController):
    """
    Class for processing all game logic. Implements functions defined in
    AbstractController that are called from UI.
    
    Attributes:
    opts (GameOptsStruct)
        Options for use in games.
    """

    def __init__(
        self,
        opts: utils.GameOptsStruct,
        *,
        notif: Optional[api.Caller] = None,
        minefield_file: Optional[os.PathLike] = None,
    ):
        """
        :param opts:
            Game options.
        :param notif:
            Optionally provide a notifier defining callbacks.
        :param minefield_file:
            Optionally provide a path to a minefield file to create the initial
            game from.
        """
        super().__init__(opts, notif=notif)

        self._drag_select = False
        self._name = ""
        self._game: game.Game
        self._last_update: _SharedInfo = _SharedInfo()
        self._notif.update_game_state(GameState.READY)
        self._notif.set_mines(self._opts.mines)
        if minefield_file:
            self.load_minefield(minefield_file)
        else:
            self.new_game()

    @property
    def board(self) -> brd.Board:
        return self._game.board

    def get_game_info(self) -> api.GameInfo:
        ret = api.GameInfo(
            game_state=self._game.state,
            x_size=self._game.x_size,
            y_size=self._game.y_size,
            mines=self._game.mines,
            per_cell=self._game.per_cell,
        )
        if self._game.state.finished():
            ret.finished_info = api.GameInfo.FinishedInfo(
                start_time=self._game.start_time,
                elapsed=self._game.get_elapsed(),
                bbbv=self._game.mf.bbbv,
                rem_bbbv=self._game.get_rem_3bv(),
                bbbvps=self._game.get_3bvps(),
                prop_complete=self._game.get_prop_complete(),
                prop_flagging=self._game.get_flag_proportion(),
            )
        return ret

    # --------------------------------------------------------------------------
    # Methods triggered by user interaction
    # --------------------------------------------------------------------------
    def new_game(self) -> None:
        """See AbstractController."""
        super().new_game()
        if self._opts.mines > self._opts.per_cell * (
            self._opts.x_size * self._opts.y_size - 1
        ):
            logger.debug(
                "Reducing number of mines from %d to %d because they don't fit",
                self._opts.mines,
                self._opts.x_size * self._opts.y_size - 1,
            )
            self._opts.mines = self._opts.x_size * self._opts.y_size - 1
            self._notif.set_mines(self._opts.mines)
        self._game = game.Game(
            x_size=self._opts.x_size,
            y_size=self._opts.y_size,
            mines=self._opts.mines,
            per_cell=self._opts.per_cell,
            lives=self._opts.lives,
            first_success=self._opts.first_success,
        )
        self._send_reset_update()

    def restart_game(self) -> None:
        """See AbstractController."""
        if not self._game.mf:
            return
        super().restart_game()
        self._game = game.Game(minefield=self._game.mf, lives=self._opts.lives)
        self._send_reset_update()

    def select_cell(self, coord: Coord_T) -> None:
        """See AbstractController."""
        super().select_cell(coord)
        cells = self._game.select_cell(coord)
        self._send_updates(cells)

    def flag_cell(self, coord: Coord_T, *, flag_only: bool = False) -> None:
        """See AbstractController."""
        super().flag_cell(coord)

        cell_state = self._game.board[coord]
        if cell_state is CellUnclicked():
            self._game.set_cell_flags(coord, 1)
        elif isinstance(cell_state, CellFlag):
            if cell_state.num == self._opts.per_cell:
                if flag_only:
                    return
                self._game.set_cell_flags(coord, 0)
            else:
                self._game.set_cell_flags(coord, cell_state.num + 1)

        self._send_updates({coord: self._game.board[coord]})

    def remove_cell_flags(self, coord: Coord_T) -> None:
        """See AbstractController."""
        super().remove_cell_flags(coord)
        self._game.set_cell_flags(coord, 0)
        self._send_updates({coord: self._game.board[coord]})

    def chord_on_cell(self, coord: Coord_T) -> None:
        """See AbstractController."""
        super().chord_on_cell(coord)
        cells = self._game.chord_on_cell(coord)
        self._send_updates(cells)

    def resize_board(self, x_size: int, y_size: int, mines: int) -> None:
        """See AbstractController."""
        super().resize_board(x_size=x_size, y_size=y_size, mines=mines)
        if (
            x_size == self._opts.x_size
            and y_size == self._opts.y_size
            and mines == self._opts.mines
        ):
            logger.info(
                "No resize required as the parameters are unchanged, starting a new game"
            )
            self.new_game()
            return

        logger.info(
            "Resizing board from %sx%s with %s mines to %sx%s with %s mines",
            self._opts.x_size,
            self._opts.y_size,
            self._opts.mines,
            x_size,
            y_size,
            mines,
        )
        self._opts.x_size = x_size
        self._opts.y_size = y_size
        self._opts.mines = mines

        self._game = game.Game(
            x_size=self._opts.x_size,
            y_size=self._opts.y_size,
            mines=self._opts.mines,
            per_cell=self._opts.per_cell,
            lives=self._opts.lives,
            first_success=self._opts.first_success,
        )
        self._send_resize_update()

    def set_first_success(self, value: bool) -> None:
        """
        Set whether the first click should be a guaranteed success.
        """
        super().set_first_success(value)
        self._opts.first_success = value
        if not self._game.state.started():
            self._game.first_success = value

    def set_per_cell(self, value: int) -> None:
        """
        Set the maximum number of mines per cell.
        """
        super().set_per_cell(value)
        if self._opts.per_cell != value:
            self._opts.per_cell = value
            if not (self._game.state.started() or self._game.minefield_known):
                self.new_game()

    def save_current_minefield(self, file: os.PathLike) -> None:
        """
        Save the current minefield to file.

        :param file:
            The location of the file to save to. Should have the extension
            ".mgb".
        """
        super().save_current_minefield(file)
        if self._game.mf is None:
            logger.warning("Unable to save current minefield - doesn't exist")
            return
        _save_minefield(self._game.mf, file)

    def load_minefield(self, file: os.PathLike) -> None:
        """
        Load a minefield from file.

        :param file:
            The location of the file to load from. Should have the extension
            ".mgb".
        """
        with open(file) as f:
            mf = brd.Minefield.from_json(json.load(f))

        self._opts.x_size = mf.x_size
        self._opts.y_size = mf.y_size
        self._opts.mines = mf.nr_mines
        self._game = game.Game(minefield=mf, lives=self._opts.lives)
        self._send_resize_update()

    # --------------------------------------------------------------------------
    # Helper methods
    # --------------------------------------------------------------------------
    def _send_reset_update(self) -> None:
        self._send_updates(dict())
        self._notif.reset()
        self._last_update = _SharedInfo()

    def _send_resize_update(self) -> None:
        self._send_updates(dict())
        self._notif.resize_minefield(self._opts.x_size, self._opts.y_size)
        self._notif.set_mines(self._opts.mines)
        self._notif.reset()

    def _send_updates(self, cells_updated: Dict[Coord_T, CellContentsType]) -> None:
        """Send updates to registered listeners."""
        end_game_info = None
        if (
            self._game.state.finished()
            and self._game.state is not self._last_update.game_state
        ):
            end_game_info = api.EndedGameInfo(
                game_state=self._game.state,
                difficulty=self._game.difficulty,
                per_cell=self._game.per_cell,
                start_time=self._game.start_time,
                elapsed=self._game.get_elapsed(),
                bbbv=self._game.mf.bbbv,
                flagging=self._game.get_flag_proportion(),
                minefield_known=self._game.minefield_known,
            )
        update = _SharedInfo(
            cell_updates=cells_updated,
            mines_remaining=self._game.mines_remaining,
            lives_remaining=self._game.lives_remaining,
            game_state=self._game.state,
            end_game_info=end_game_info,
        )

        # Send updates to registered frontends.
        if update.cell_updates:
            self._notif.update_cells(update.cell_updates)
        if update.mines_remaining != self._last_update.mines_remaining:
            self._notif.update_mines_remaining(update.mines_remaining)
        # if update.lives_remaining != self._last_update.lives_remaining:
        #     self._notif.update_lives_remaining(update.lives_remaining)
        if update.game_state is not self._last_update.game_state:
            self._notif.update_game_state(update.game_state)
        if update.end_game_info is not None:
            self._notif.handle_finished_game(update.end_game_info)

        self._last_update = update


class CreateController(api.AbstractController):
    """A controller for creating boards."""

    def __init__(
        self, opts: utils.GameOptsStruct, *, notif: Optional[api.Caller] = None
    ):
        super().__init__(opts, notif=notif)
        self._board: brd.Board = brd.Board(self._opts.x_size, self._opts.y_size)
        self._flags: int = 0
        self._notif.update_game_state(GameState.READY)
        self._notif.set_mines(self._flags)
        self._notif.reset()

    @property
    def board(self) -> brd.Board:
        return self._board

    def get_game_info(self) -> api.GameInfo:
        return api.GameInfo(
            game_state=GameState.ACTIVE,
            x_size=self._opts.x_size,
            y_size=self._opts.y_size,
            mines=self._flags,
            per_cell=self._opts.per_cell,
        )

    def new_game(self) -> None:
        """See AbstractController."""
        super().new_game()
        self._board = brd.Board(self._opts.x_size, self._opts.y_size)
        self._flags = 0
        self._notif.reset()

    def restart_game(self) -> None:
        super().restart_game()
        self.new_game()

    def select_cell(self, coord: Coord_T) -> None:
        super().select_cell(coord)
        cell = self._board[coord]
        if cell is CellUnclicked():
            self._board[coord] = CellNum(0)
        elif isinstance(cell, CellNum):
            self._board[coord] += 1
        else:
            return
        self._notif.update_cells({coord: self._board[coord]})

    def flag_cell(self, coord: Coord_T, *, flag_only: bool = False) -> None:
        super().flag_cell(coord)
        cell = self._board[coord]

        if cell is CellUnclicked():
            self._board[coord] = CellMine(1)
            self._flags += 1
        elif isinstance(cell, CellMine):
            if cell.num == self._opts.per_cell:
                if flag_only:
                    return
                self._board[coord] = CellUnclicked()
                self._flags -= self._opts.per_cell
            else:
                self._board[coord] += 1
                self._flags += 1
        elif isinstance(cell, CellNum):
            self.board[coord] = CellUnclicked()
        else:
            return
        self._notif.update_cells({coord: self._board[coord]})
        self._notif.update_mines_remaining(self._flags)

    def chord_on_cell(self, coord: Coord_T) -> None:
        super().chord_on_cell(coord)

    def remove_cell_flags(self, coord: Coord_T) -> None:
        super().remove_cell_flags(coord)
        cell = self._board[coord]
        if not isinstance(cell, CellMine):
            return
        self._board[coord] = CellUnclicked()
        self._flags -= cell.num
        self._notif.update_cells({coord: self._board[coord]})
        self._notif.update_mines_remaining(self._flags)

    def resize_board(self, x_size: int, y_size: int, mines: int) -> None:
        super().resize_board(x_size=x_size, y_size=y_size, mines=mines)
        if x_size == self._opts.x_size and y_size == self._opts.y_size:
            logger.info(
                "No resize required as the parameters are unchanged, starting a new game"
            )
            self.new_game()
            return

        logger.info(
            "Resizing board from %sx%s to %sx%s",
            self._opts.x_size,
            self._opts.y_size,
            x_size,
            y_size,
        )
        self._opts.x_size = x_size
        self._opts.y_size = y_size
        self._notif.resize_minefield(x_size, y_size)
        self.new_game()

    def set_first_success(self, value: bool) -> None:
        super().set_first_success(value)
        # Store the value so it can be retrieved from the next controller.
        self._opts.first_success = value

    def set_per_cell(self, value: int) -> None:
        super().set_per_cell(value)
        self._opts.per_cell = value

    def save_current_minefield(self, file: os.PathLike) -> None:
        """
        Save the current created minefield to file.

        :param file:
            The location of the file to save to. Should have the extension
            ".mgb".
        """
        super().save_current_minefield(file)
        mines_grid = utils.Grid(self._opts.x_size, self._opts.y_size)
        for c in self._board.all_coords:
            if type(self._board[c]) is CellMine:
                mines_grid[c] = self._board[c].num
        mf = brd.Minefield.from_grid(mines_grid, per_cell=self._opts.per_cell)
        _save_minefield(mf, file)

    def load_minefield(self, file: os.PathLike) -> None:
        return NotImplemented



"""
__init__.py - Available imports from the package

April 2018, Lewis Gaul
"""

__all__ = (
    "BaseController",
    "CreateController",
    "GameController",
    "Minefield",
    "api",
    "board",
    "engine",
    "game",
)

from . import api, board, engine, game
from .board import Minefield
from .engine import BaseController, CreateController, GameController



"""
game.py - Game logic

March 2018, Lewis Gaul

Exports:
Game (class)
    Representation of a minesweeper game.
"""

__all__ = ("Game", "GameNotStartedError")

import functools
import logging
import math
import time as tm
from typing import Callable, Dict, Iterable, Optional, Tuple, Union

from ..shared import utils
from ..types import (
    CellContentsType,
    CellFlag,
    CellHitMine,
    CellMine,
    CellMineType,
    CellNum,
    CellUnclicked,
    CellWrongFlag,
    GameState,
)
from ..typing import Coord_T
from .board import Board, Minefield


logger = logging.getLogger(__name__)


def _check_coord(method: Callable) -> Callable:
    """
    Wrap a method that takes a coord to check it is inside the valid range.

    :raise ValueError:
        If the coord is not valid.
    """

    @functools.wraps(method)
    def wrapped(self: "Game", coord: Coord_T, *args, **kwargs):
        if not 0 <= coord[0] < self.x_size or not 0 <= coord[1] < self.y_size:
            raise ValueError(
                f"Coordinate is out of bounds, should be between (0,0) and "
                f"({self.x_size-1}, {self.y_size-1})"
            )
        return method(self, coord, *args, **kwargs)

    return wrapped


def _ignore_if(
    *,
    game_state: Optional[
        Union[str, GameState, Iterable[str], Iterable[GameState]]
    ] = None,
    cell_state: Optional[Union[type, Tuple[type, ...]]] = None,
) -> Callable:
    """
    Return a decorator which prevents a method from running if any of the given
    parameters are satisfied.

    Arguments:
    game_state=None (GameState | (GameState, ...) | None)
        A game state or iterable of game states to match against.
    cell_state=None (subclass of CellContentsType | (..., ...) | None)
        A cell contents type or iterable of the same to match against. The
        decorated method must take the cell coordinate as the first argument.
    """

    def decorator(method: Callable) -> Callable:
        # If the cell state is specified it is assumed the coord is passed in
        # as the first arg.
        if cell_state is None:

            @functools.wraps(method)
            def wrapped(self: "Game", *args, **kwargs):
                if game_state is not None and (
                    self.state == game_state or self.state in game_state
                ):
                    return
                return method(self, *args, **kwargs)

        else:

            @functools.wraps(method)
            def wrapped(self: "Game", coord: Coord_T, *args, **kwargs):
                if (
                    game_state is not None
                    and (self.state == game_state or self.state in game_state)
                ) or isinstance(self.board[coord], cell_state):
                    return
                return method(self, coord, *args, **kwargs)

        return wrapped

    return decorator


def _ignore_if_not(
    *,
    game_state: Optional[
        Union[str, GameState, Iterable[str], Iterable[GameState]]
    ] = None,
    cell_state: Optional[Union[type, Tuple[type, ...]]] = None,
) -> Callable:
    """
    Return a decorator which prevents a method from running if any of the given
    parameters are satisfied.

    Arguments:
    game_state=None (GameState | (GameState, ...) | None)
        A game state or iterable of game states to match against.
    cell_state=None (subclass of CellContentsType | (..., ...) | None)
        A cell contents type or iterable of the same to match against. The
        decorated method must take the cell coordinate as the first argument.
    """

    def decorator(method: Callable) -> Callable:
        # If the cell state is specified it is assumed the coord is passed in
        # as the first arg.
        if cell_state is None:

            @functools.wraps(method)
            def wrapped(self: "Game", *args, **kwargs):
                if (
                    game_state is not None
                    and self.state != game_state
                    and self.state not in game_state
                ):
                    return
                return method(self, *args, **kwargs)

        else:

            @functools.wraps(method)
            def wrapped(self: "Game", coord: Coord_T, *args, **kwargs):
                if (
                    game_state is not None
                    and self.state != game_state
                    and self.state not in game_state
                ) or not isinstance(self.board[coord], cell_state):
                    return
                return method(self, coord, *args, **kwargs)

        return wrapped

    return decorator


def _check_game_started(method: Callable) -> Callable:
    """Check the game has been started, raising an error if not."""

    @functools.wraps(method)
    def wrapped(self: "Game", *args, **kwargs):
        if self.state is GameState.READY:
            raise GameNotStartedError("Minefield not yet created")
        assert self.mf is not None
        assert self.start_time is not None
        return method(self, *args, **kwargs)

    return wrapped


class GameNotStartedError(Exception):
    """Game has not been started, so no minefield has been created."""


class Game:
    """
    A minesweeper game, storing a minefield and the state of a game, including
    the board and other game settings. Provides methods to start the game and
    standard interactions such as selecting or flagging a cell and chording.
    """

    def __init__(
        self,
        *,
        x_size: Optional[int] = None,
        y_size: Optional[int] = None,
        mines: Optional[int] = None,
        per_cell: int = 1,
        lives: int = 1,
        first_success: bool = False,
        minefield: Optional[Minefield] = None,
    ):
        """
        :param x_size:
            Number of columns in the grid. Ignored if a minefield is passed in,
            since the minefield already has a size.
        :param y_size:
            Number of rows in the grid. Ignored if a minefield is passed in, since
            the minefield already has a size.
        :param mines:
            The number of mines. Ignored if a minefield is passed in, since the
            minefield should already contain a number of mines.
        :param per_cell:
            Maximum number of mines per cell. Ignored if a minefield is passed in,
            since the minefield should already refer to a max per cell.
        :param lives:
            A number of lives available during the game.
        :param first_success:
            Whether the first cell selected should be guaranteed to be safe, and
            give an opening if possible. Ignored if a minefield is passed in, as
            the minefield is not created in response to the first select, so it
            is not possible to guarantee success on the first select.
        :param minefield:
            A minefield to use for the game. Takes precedence over various other
            arguments, see above.
        :raise ValueError:
            If the number of mines is too high to fit in the grid.
        """
        self.mf: Optional[Minefield]
        self.minefield_known: bool
        if minefield:
            x_size, y_size = minefield.x_size, minefield.y_size
            mines = minefield.nr_mines
            per_cell = minefield.per_cell
            first_success = False
            self.mf = minefield
            self.minefield_known = True
        else:
            if x_size is None or y_size is None or mines is None:
                raise ValueError(
                    "x_size, y_size and mines must be integers if a minefield is not provided"
                )
            Minefield.check_enough_space(
                x_size=x_size, y_size=y_size, mines=mines, per_cell=per_cell
            )
            self.mf = None
            self.minefield_known = False
        self.x_size: int = x_size
        self.y_size: int = y_size
        self.mines: int = mines
        self.per_cell: int = per_cell
        self.lives: int = lives
        self.first_success: bool = first_success
        self.board: Board = Board(x_size, y_size)
        self.start_time: Optional[float] = None
        self.end_time: Optional[float] = None
        self.state: GameState = GameState.READY
        self.mines_remaining: int = self.mines
        self.lives_remaining: int = self.lives
        self._cell_updates: Dict[Coord_T, CellContentsType] = dict()
        self._num_flags: int = 0

    @property
    def difficulty(self) -> str:
        return utils.get_difficulty(self.x_size, self.y_size, self.mines)

    @_check_game_started
    def get_rem_3bv(self) -> int:
        """
        Calculate the minimum remaining number of clicks needed to solve.
        """
        if self.state == GameState.WON:
            return 0
        else:
            lost_mf = Minefield.from_grid(self.mf, per_cell=self.per_cell)
            # Replace any openings already found with normal clicks (ones).
            for c in self.board.all_coords:
                if type(self.board[c]) is CellNum:
                    lost_mf.completed_board[c] = CellNum(1)
            # Find the openings which remain.
            lost_mf.openings = lost_mf._find_openings()
            rem_opening_coords = {c for opening in lost_mf.openings for c in opening}
            # Count the number of essential clicks that have already been
            # done by counting clicked cells minus the ones at the edge of
            # an undiscovered opening.
            completed_3bv = len(
                {c for c in self.board.all_coords if type(self.board[c]) is CellNum}
                - rem_opening_coords
            )
            return lost_mf._calc_3bv() - completed_3bv

    @_check_game_started
    def get_prop_complete(self) -> float:
        """
        Calculate the progress of solving the board using 3bv.
        """
        return (self.mf.bbbv - self.get_rem_3bv()) / self.mf.bbbv

    @_check_game_started
    def get_3bvps(self) -> float:
        """
        Calculate the 3bv/s based on current progress.

        :return:
            The 3bv/s, or math.inf if elapsed time is zero.
        """
        if self.get_elapsed() == 0:
            return math.inf
        return self.mf.bbbv * self.get_prop_complete() / self.get_elapsed()

    def get_elapsed(self) -> float:
        """
        Get the elapsed game time.

        Returns zero if the game has not been started.

        :return:
            The elapsed time, in seconds.
        """
        if self.end_time:
            return self.end_time - self.start_time
        elif self.start_time:
            return tm.time() - self.start_time
        else:
            return 0

    def get_flag_proportion(self) -> float:
        if self.mines == 0:
            return 0
        return self._num_flags / self.mines

    def _create_minefield(self, coord: Coord_T) -> None:
        """Create the minefield in response to a cell being selected."""
        if self.first_success:
            safe_coords = self.board.get_nbrs(coord, include_origin=True)
            logger.debug(
                "Trying to create minefield with the following safe coordinates: %s",
                safe_coords,
            )
            try:
                self.mf = Minefield(
                    self.x_size,
                    self.y_size,
                    mines=self.mines,
                    per_cell=self.per_cell,
                    safe_coords=safe_coords,
                )
            except ValueError:
                logger.info(
                    "Unable to give opening on the first click, "
                    "still ensuring a safe click"
                )
                # This should be guaranteed to succeed.
                self.mf = Minefield(
                    self.x_size,
                    self.y_size,
                    mines=self.mines,
                    per_cell=self.per_cell,
                    safe_coords=[coord],
                )
            else:
                logger.debug("Successfully created minefield")
        else:
            logger.debug("Creating minefield without guaranteed first click success")
            self.mf = Minefield(
                self.x_size, self.y_size, mines=self.mines, per_cell=self.per_cell
            )

    def _set_cell(self, coord: Coord_T, state: CellContentsType):
        """
        Set the contents of a cell and store the update.

        :param coord:
            The coordinate of the cell to set.
        :param state:
            The state to set the cell to.
        """
        self.board[coord] = state
        self._cell_updates[coord] = state

    def _select_cell_action(self, coord: Coord_T) -> None:
        """
        Implementation of the action of selecting/clicking a cell.
        """
        if self.mf.cell_contains_mine(coord):
            logger.debug("Mine hit at %s", coord)
            self._set_cell(coord, CellHitMine(self.mf[coord]))
            self.lives_remaining -= 1

            if self.lives_remaining == 0:
                logger.info("Game lost")
                self.end_time = tm.time()
                self.state = GameState.LOST

                for c in self.mf.all_coords:
                    if (
                        self.mf.cell_contains_mine(c)
                        and self.board[c] is CellUnclicked()
                    ):
                        self._set_cell(c, CellMine(self.mf[c]))

                    elif (
                        type(self.board[c]) is CellFlag
                        and self.board[c] != self.mf.completed_board[c]
                    ):
                        self._set_cell(c, CellWrongFlag(self.board[c].num))
            else:
                self.mines_remaining -= self.mf[coord]
        elif self.mf.completed_board[coord] is CellNum(0):
            for full_opening in self.mf.openings:
                if coord in full_opening:
                    # Found the opening, quit the loop here.
                    logger.debug("Opening hit: %s", full_opening)
                    break
            else:
                logger.error(
                    "Coordinate %s not found in openings %s", coord, self.mf.openings
                )

            # Get the propagation of cells forming part of the opening.
            opening = set()  # Coords belonging to the opening
            check = {coord}  # Coords whose neighbours need checking
            while check:
                c = check.pop()
                unclicked_nbrs = {
                    z
                    for z in self.board.get_nbrs(c, include_origin=True)
                    if self.board[z] is CellUnclicked()
                }
                check |= {
                    z
                    for z in unclicked_nbrs - opening
                    if self.mf.completed_board[z] is CellNum(0)
                }
                opening |= unclicked_nbrs

            logger.debug("Propagated opening: %s", list(opening))
            bad_opening_cells = {}
            for c in opening:
                if self.board[c] is CellUnclicked():
                    self._set_cell(c, self.mf.completed_board[c])
                else:
                    bad_opening_cells[c] = self.board[c]
            if bad_opening_cells:
                logger.error(
                    "Should only have clicked cells in opening, found: %s",
                    bad_opening_cells,
                )
        else:
            logger.debug("Regular cell revealed")
            self._set_cell(coord, self.mf.completed_board[coord])

    def _check_for_completion(self) -> None:
        """
        Check if game is complete by comparing the board to the minefield's
        completed board. If it is, display flags in remaining unclicked cells.
        """
        # Assume (for contradiction) that game is complete.
        is_complete = True
        for c in self.mf.all_coords:
            exp_val = self.mf.completed_board[c]
            if type(exp_val) is CellNum and exp_val != self.board[c]:
                is_complete = False
                break

        if is_complete:
            logger.info("Game won")

            self.end_time = tm.time()
            self.state = GameState.WON
            self.mines_remaining = 0

            for c in self.mf.all_coords:
                if (
                    self.mf.cell_contains_mine(c)
                    and type(self.board[c]) is not CellHitMine
                ):
                    self._set_cell(c, CellFlag(self.mf[c]))

    @_check_coord
    @_ignore_if_not(game_state=("READY", "ACTIVE"), cell_state=CellUnclicked)
    def select_cell(self, coord: Coord_T) -> Dict[Coord_T, CellContentsType]:
        """
        Perform the action of selecting/clicking a cell. Game must be started
        before calling this method.
        """
        just_started = False
        if self.state == GameState.READY:
            if not self.mf:
                self._create_minefield(coord)
            self.state = GameState.ACTIVE
            self.start_time = tm.time()
            just_started = True
        self._select_cell_action(coord)
        if not self.state.finished():
            self._check_for_completion()
            if self.state is GameState.WON and just_started:
                self.end_time = self.start_time
        try:
            return self._cell_updates
        finally:
            self._cell_updates = dict()

    @_check_coord
    @_ignore_if_not(
        game_state=("READY", "ACTIVE"), cell_state=(CellFlag, CellUnclicked)
    )
    def set_cell_flags(
        self, coord: Coord_T, nr_flags: int
    ) -> Dict[Coord_T, CellContentsType]:
        """Set the number of flags in a cell."""
        if nr_flags < 0 or nr_flags > self.per_cell:
            raise ValueError(
                f"Invalid number of flags ({nr_flags}) - should be between 0 and "
                f"{self.per_cell}"
            )

        old_nr_flags = (
            0 if self.board[coord] is CellUnclicked() else self.board[coord].num
        )
        if nr_flags == 0:
            self._set_cell(coord, CellUnclicked())
        else:
            self._set_cell(coord, CellFlag(nr_flags))
        self.mines_remaining += old_nr_flags - nr_flags
        self._num_flags += nr_flags - old_nr_flags

        try:
            return self._cell_updates
        finally:
            self._cell_updates = dict()

    @_check_coord
    @_ignore_if_not(game_state="ACTIVE", cell_state=CellNum)
    def chord_on_cell(self, coord: Coord_T) -> Dict[Coord_T, CellContentsType]:
        """Chord on a cell that contains a revealed number."""
        nbrs = self.board.get_nbrs(coord)
        num_flagged_nbrs = sum(
            [self.board[c].num for c in nbrs if isinstance(self.board[c], CellMineType)]
        )
        logger.debug(
            "%s flagged mine(s) around clicked cell showing number %s",
            num_flagged_nbrs,
            self.board[coord],
        )

        unclicked_nbrs = [c for c in nbrs if self.board[c] is CellUnclicked()]
        if self.board[coord] != CellNum(num_flagged_nbrs) or not unclicked_nbrs:
            return dict()

        logger.info("Successful chording, selecting cells %s", unclicked_nbrs)
        for c in unclicked_nbrs:
            self._select_cell_action(c)

        if self.state != GameState.LOST:
            self._check_for_completion()

        try:
            return self._cell_updates
        finally:
            self._cell_updates = dict()



"""
__init__.py - Bot package

February 2020, Lewis Gaul
"""

__all__ = ("init_route_handling",)

import json
import logging
import os
import sys

import flask

from minegauler.shared import highscores as hs

from .. import add_new_highscore_hook
from ..utils import is_highscore_new_best
from . import routes, utils


logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# External
# ------------------------------------------------------------------------------


def init_route_handling(app: flask.app.Flask):
    if "BOT_ACCESS_TOKEN" not in os.environ:
        logger.error("No 'BOT_ACCESS_TOKEN' env var set")
        sys.exit(1)
    utils.set_bot_access_token(os.environ["BOT_ACCESS_TOKEN"])

    try:
        with open(utils.USER_NAMES_FILE) as f:
            utils.USER_NAMES = json.load(f)
    except FileNotFoundError:
        logger.warning("%s file not found", utils.USER_NAMES_FILE)

    routes.activate_bot_msg_handling(app)

    add_new_highscore_hook(_new_highscore_hook)


# ------------------------------------------------------------------------------
# Internal
# ------------------------------------------------------------------------------


def _new_highscore_hook(highscore: hs.HighscoreStruct) -> None:
    if highscore.name != "Siwel G":
        try:
            utils.send_myself_message(f"New highscore added:\n{highscore}")
        except Exception:
            logger.exception("Error sending webex message")

    if is_highscore_new_best(highscore) == "time":
        try:
            utils.send_new_best_message(highscore)
        except Exception:
            logger.exception("Error sending webex message for new best")



"""
__main__.py - Entry point for the application.

December 2018, Lewis Gaul
"""

# Workaround to get the mysql libcrypto and libssl libs loaded.
import mysql.connector  # isort:skip

import logging
import sys

from . import core, frontend, shared
from ._version import __version__


logger = logging.getLogger(__name__)

logging.basicConfig(
    filename="runtime.log",
    level=logging.DEBUG,
    format="%(asctime)s[%(levelname)s](%(name)s) %(message)s",
)


read_settings = shared.read_settings_from_file()

if read_settings:
    game_opts = shared.GameOptsStruct.from_structs(read_settings)
    gui_opts = shared.GUIOptsStruct.from_structs(read_settings)
    logger.info("Settings read from file")
else:
    logger.info("Using default settings")
    game_opts = shared.GameOptsStruct()
    gui_opts = shared.GUIOptsStruct()
logger.debug("Game options: %s", game_opts)
logger.debug("GUI options: %s", gui_opts)


logger.info("Starting up")

# Create core controller.
ctrlr = core.BaseController(game_opts)
# Init frontend and create controller.
frontend.init_app()
gui = frontend.MinegaulerGUI(ctrlr, frontend.state.State.from_opts(game_opts, gui_opts))
# Register frontend with core controller.
ctrlr.register_listener(gui)
# Perform initialisation again now that the frontend is registered.
ctrlr.reinitialise()

# Run the app.
logger.debug("Entering event loop")
rc = frontend.run_app(gui)
logger.debug("Exiting event loop")


persist_settings = shared.AllOptsStruct.from_structs(
    ctrlr.get_game_options(), gui.get_gui_opts()
)
shared.write_settings_to_file(persist_settings)


logger.info("Exiting with exit code %d", rc)

sys.exit(rc)



"""
utils.py - Utilities for the frontend.

December 2018, Lewis Gaul
"""

__all__ = ("FILES_DIR", "IMG_DIR")

import pathlib

from minegauler import ROOT_DIR


IMG_DIR: pathlib.Path = ROOT_DIR / "images"
FILES_DIR: pathlib.Path = ROOT_DIR / "files"
